/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "scrnintstr.h"
#include "resource.h"
#include "dixstruct.h"
#include "../../fb/fb.h"

#include "Agent.h"
#include "Display.h"
#include "Visual.h"
#include "Drawable.h"
#include "Pixmap.h"
#include "GC.h"
#include "Image.h"
#include "Events.h"
#include "Client.h"
#include "Trap.h"
#include "Split.h"
#include "Args.h"
#include "Screen.h"

#include "NXlib.h"
#include "NXpack.h"

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG
#undef  DUMP

/*
 * Don't try to pack images having a data size
 * smaller than this threshold.
 */

#define IMAGE_PACK_THRESHOLD  768

/*
 * Don't clip images smaller than this.
 */

#define IMAGE_CLIP_THRESHOLD  16384

/*
 * Introduce a small delay after each image
 * operation if the session is down. Value
 * is in microseconds and is multiplied by
 * the image data size in kilobytes.
 */

#define IMAGE_DELAY_IF_DOWN  250

/*
 * Time in milliseconds after which a corrupted
 * drawable is considered too much old and must
 * be synchronized.
 */
/*
FIXME: Should we set a dynamic aging based
       on link type?
*/
#define MAX_DRAWABLE_CORRUPTED_AGING  2000

/*
 * Set if images can use the alpha channel.
 */

int nxagentAlphaEnabled = 0;
int nxagentAlphaCompat  = 0;
int nxagentAlphaPacked  = 0;

/*
 * Used to reformat image when connecting to
 * displays having different byte order.
 */

extern void BitOrderInvert(unsigned char *, int);
extern void TwoByteSwap(unsigned char *, register int);
extern void FourByteSwap(register unsigned char *, register int);

/*
 * Store the last visual used to unpack
 * the images for the given client.
 */

static VisualID nxagentUnpackVisualId[MAX_CONNECTIONS];

/*
 * Store the last alpha data set for the
 * client.
 */

typedef struct _UnpackAlpha
{
  char *data;
  int size;

} UnpackAlphaRec;

typedef UnpackAlphaRec *UnpackAlphaPtr;

static UnpackAlphaPtr nxagentUnpackAlpha[MAX_CONNECTIONS];

/*
 * Encode the imade alpha channel by using
 * a specific encoding, separating it from
 * the rest of the RGB data.
 */

static char *nxagentImageAlpha(XImage *ximage);
static void nxagentSetUnpackAlpha(DrawablePtr pDrawable, XImage *pImage, ClientPtr pClient);

/*
 * Reset the visual and alpha cache before
 * closing the screen or connecting to a
 * different display.
 */

void nxagentResetVisualCache(void);
void nxagentResetAlphaCache(void);

/*
 * Collect the image cache statistics.
 */

typedef struct _ImageStatisticsRec
{
  double partialEncoded;
  double partialLookups;
  double partialAdded;
  double partialFound;

  double totalEncoded;
  double totalLookups;
  double totalAdded;
  double totalFound;

} ImageStatisticsRec;

ImageStatisticsRec nxagentImageStatistics;

int nxagentImageReformat(char *base, int nbytes, int bpp, int order)
{
  /*
   * This is used whenever we need to swap the image data.
   * If we got an image from a X server having a different
   * endianess, we will need to redormat the image to match
   * our own image-order so that ProcGetImage can return
   * the expected format to the client.
   */

  switch (bpp)
  {
    case 1:
    {
      if (BITMAP_BIT_ORDER != order)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentImageReformat: Bit order invert with size [%d] "
                    "bits per pixel [%d] byte order [%d].\n", nbytes, bpp, order);
        #endif

        BitOrderInvert((unsigned char *) base, nbytes);
      }

      #if IMAGE_BYTE_ORDER != BITMAP_BIT_ORDER && BITMAP_SCANLINE_UNIT != 8

      nxagentImageReformat(base, nbytes, BITMAP_SCANLINE_UNIT, order);

      #endif

      break;
    }
    case 4:
    case 8:
    {
      break;
    }
    case 16:
    {
      if (IMAGE_BYTE_ORDER != order)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentImageReformat: Two bytes swap with size [%d] "
                    "bits per pixel [%d] byte order [%d].\n", nbytes, bpp, order);
        #endif

        TwoByteSwap((unsigned char *) base, nbytes);
      }

      break;
    }
    case 32:
    {
      if (IMAGE_BYTE_ORDER != order)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentImageReformat: Four bytes swap with size [%d] "
                    "bits per pixel [%d] byte order [%d].\n", nbytes, bpp, order);
        #endif

        FourByteSwap((unsigned char *) base, nbytes);
      }

      break;
    }
  }

  return 1;
}

int nxagentImageLength(int width, int height, int format, int leftPad, int depth)
{
  int line = 0;

  if (format == XYBitmap)
  {
    line = BitmapBytePad(width + leftPad);
  }
  else if (format == XYPixmap)
  {
    line  = BitmapBytePad(width + leftPad);

    line *= depth;
  }
  else if (format == ZPixmap)
  {
    line = PixmapBytePad(width, depth);
  }

  return line * height;
}

int nxagentImagePad(int width, int format, int leftPad, int depth)
{
  int line = 0;

  if (format == XYBitmap)
  {
    line = BitmapBytePad(width + leftPad);
  }
  else if (format == XYPixmap)
  {
    line = BitmapBytePad(width + leftPad);
  }
  else if (format == ZPixmap)
  {
    line = PixmapBytePad(width, depth);
  }

  return line;
}

char *nxagentImageAlpha(XImage *image)
{
  char *pData;

  char *pSrcData;
  char *pDstData;

  int size;
  int offset;

  /*
   * Use one byte per pixel.
   */

  size = (image -> bytes_per_line * image -> height) >> 2;

  pData = Xmalloc(size);

  if (pData == NULL)
  {
    return NULL;
  }

  /*
   * The image is supposed to be in
   * server order.
   */

  offset = (image -> byte_order == MSBFirst) ? 0 : 3;

  pSrcData = image -> data;
  pDstData = pData;

  while (size-- > 0)
  {
    *pDstData++ = *(pSrcData + offset);

    pSrcData += 4;
  }

  return pData;
}

/*
 * Write down the image cache statistics
 * to the buffer.
 */

void nxagentImageStatisticsHandler(char **buffer, int type)
{
/*
FIXME: Agent cache statistics have to be implemented.
*/
  #ifdef TEST
  fprintf(stderr, "nxagentImageStatisticsHandler: STATISTICS! Statistics requested to the agent.\n");
  #endif

  *buffer = NULL;
}

/*
 * This should be called only for drawables
 * having a depth of 32. In the other cases,
 * it would only generate useless traffic.
 */

void nxagentSetUnpackAlpha(DrawablePtr pDrawable, XImage *pImage, ClientPtr pClient)
{
  int resource = pClient -> index;

  unsigned int size = (pImage -> bytes_per_line * pImage -> height) >> 2;

  char *data = nxagentImageAlpha(pImage);

  if (data == NULL)
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentSetUnpackAlpha: PANIC! Can't allocate data for the alpha channel.\n");
    #endif

    return;
  }

  /*
   * If we are synchronizing the drawable, discard
   * any unpack alpha stored for the client. The
   * alpha data, in fact, may be still traveling
   * and so we either wait until the end of the
   * split or send a fresh copy.
   */
/*
FIXME: Here the split trap is always set and so the caching of
       the alpha channel is useless. I remember we set the trap
       because of the cursor but why is it always set now?
*/
  #ifdef DEBUG
  fprintf(stderr, "nxagentSetUnpackAlpha: Checking alpha channel for client [%d] with trap [%d].\n",
              resource, nxagentSplitTrap);
  #endif

  if (nxagentSplitTrap == 1 || nxagentUnpackAlpha[resource] == NULL ||
          nxagentUnpackAlpha[resource] -> size != size ||
              memcmp(nxagentUnpackAlpha[resource] -> data, data, size) != 0)
  {
    #ifdef DEBUG
    fprintf(stderr, "nxagentSetUnpackAlpha: Sending alpha channel with width [%d] height [%d] "
                "bytes per line [%d] size [%d].\n", pImage -> width, pImage -> height,
                    pImage -> bytes_per_line, size);
    #endif

    /*
     * Check if we are connected to a newer proxy
     * version and so can send the alpha data in
     * compressed form.
     */

    if (nxagentAlphaCompat == 0)
    {
      unsigned int packedSize = 0;
      char *packedData;

      if (nxagentAlphaPacked == 1)
      {
        #ifdef DEBUG
        fprintf(stderr, "nxagentSetUnpackAlpha: Compressing the alpha data for client [%d].\n",
                    resource);
        #endif

        packedData = NXEncodeAlpha(data, size, &packedSize);
      }
      else
      {
        packedData = NULL;
      }

      if (packedData != NULL)
      {
        NXSetUnpackAlpha(nxagentDisplay, resource, PACK_ALPHA, size, packedData, packedSize);

        Xfree(packedData);
      }
      else
      {
        NXSetUnpackAlpha(nxagentDisplay, resource, NO_PACK, size, data, size);
      }
    }
    else
    {
      NXSetUnpackAlphaCompat(nxagentDisplay, resource, size, data);
    }

    if (nxagentUnpackAlpha[resource] != NULL)
    {
      Xfree(nxagentUnpackAlpha[resource] -> data);
    }
    else if ((nxagentUnpackAlpha[resource] = Xmalloc(sizeof(UnpackAlphaRec))) == NULL)
    {
      #ifdef PANIC
      fprintf(stderr, "nxagentSetUnpackAlpha: PANIC! Can't allocate data for the alpha structure.\n");
      #endif

      Xfree(data);

      return;
    }

    #ifdef DEBUG
    fprintf(stderr, "nxagentSetUnpackAlpha: Saved alpha channel for client [%d] with size [%d].\n",
                resource, size);
    #endif

    nxagentUnpackAlpha[resource] -> size = size;
    nxagentUnpackAlpha[resource] -> data = data;
  }
  else
  {
    #ifdef DEBUG
    fprintf(stderr, "nxagentSetUnpackAlpha: Matched alpha channel for client [%d] with size [%d].\n",
                resource, size);
    #endif

    Xfree(data);
  }
}

/*
 * The NX agent's implementation of the
 * X server's image functions.
 */

void nxagentPutImage(DrawablePtr pDrawable, GCPtr pGC, int depth,
                         int dstX, int dstY, int dstWidth, int dstHeight,
                             int leftPad, int format, char *data)
{
  int length;

  RegionPtr pRegion = NullRegion;

  int resource = 0;
  int split    = 0;
  int cache    = 1;

  Time drawableAging;

  #ifdef TEST
  fprintf(stderr, "nxagentPutImage: Image data at [%p] drawable [%s][%p] geometry [%d,%d,%d,%d].\n",
              data, (pDrawable -> type == DRAWABLE_PIXMAP ? "Pixmap" : "Window"),
                  (void *) pDrawable, dstX, dstY, dstWidth, dstHeight);
  #endif

  /*
   * If the display is down, sleep for a while
   * but still give a chance to the client to
   * write to the framebuffer.
   */

  length = nxagentImageLength(dstWidth, dstHeight, format, leftPad, depth);

  if (NXQueryDisplayError(nxagentDisplay) == 1)
  {
    int us;

    us = IMAGE_DELAY_IF_DOWN * (length / 1024);

    us = (us < 10000 ? 10000 : (us > 1000000 ? 1000000 : us));

    #ifdef DEBUG
    fprintf(stderr, "nxagentPutImage: Sleeping for [%d] milliseconds with length [%d] and link down.\n",
                us / 1000, length);
    #endif

    usleep(us);
  }

  /*
   * This is of little use because clients usually write
   * to windows only after an expose event, and, in the
   * rare case they use a direct put image to the window
   * (for a media player it should be a necessity), they
   * are likely to monitor the visibility of the window.
   */

  if (pDrawable -> type == DRAWABLE_WINDOW &&
          nxagentWindowIsVisible((WindowPtr) pDrawable) == 0)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentPutImage: WARNING! Prevented operation on fully obscured "
                "window at [%p].\n", (void *) pDrawable);
    #endif

    goto nxagentPutImageEnd;
  }

  /*
   * This is more interesting. Check if the operation
   * will produce a visible result based on the clip
   * list of the window and the GC.
   */

  pRegion = nxagentCreateRegion(pDrawable, pGC, dstX, dstY, dstWidth, dstHeight);

  if (REGION_NIL(pRegion) == 1)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentPutImage: WARNING! Prevented operation on fully clipped "
                "region [%d,%d,%d,%d] for drawable at [%p].\n", pRegion -> extents.x1,
                    pRegion -> extents.y1, pRegion -> extents.x2, pRegion -> extents.y2,
                        (void *) pDrawable);
    #endif

    goto nxagentPutImageEnd;
  }

/*
FIXME: We should avoid to write the images generated by the
       GLX and the XVideo extension to the frame-buffer, at
       least when the lazy encoding is off (and so we don't
       need to generate deferred updates), or when the link
       is set to LAN.
*/
/*
FIXME: This condition looks suspicious. If we have set the
       GC trap, we should probably skip the operation on
       the real display unconditionally, and only write to
       the frame-buffer.
*/
/*
FIXME: Should use these.

  int framebuffer = 1;
  int realize     = 1;
*/
  if (nxagentGCTrap == 1 && nxagentReconnectTrap == 0 &&
          nxagentFBTrap == 0 && nxagentShmTrap == 0)
  {
    if ((pDrawable) -> type == DRAWABLE_PIXMAP)
    {
      fbPutImage(nxagentVirtualDrawable(pDrawable), pGC, depth,
                     dstX, dstY, dstWidth, dstHeight, leftPad, format, data);
    }
    else
    {
      fbPutImage(pDrawable, pGC, depth,
                     dstX, dstY, dstWidth, dstHeight, leftPad, format, data);
    }

    goto nxagentPutImageEnd;
  }

  if (pDrawable -> type == DRAWABLE_PIXMAP &&
          nxagentReconnectTrap == 0 && nxagentFBTrap == 0 &&
              nxagentShmTrap == 0)
  {
    fbPutImage(nxagentVirtualDrawable(pDrawable), pGC, depth,
                   dstX, dstY, dstWidth, dstHeight, leftPad, format, data);
  }
  else if (pDrawable -> type == DRAWABLE_WINDOW &&
               nxagentReconnectTrap == 0)
  {
    fbPutImage(pDrawable, pGC, depth,
                   dstX, dstY, dstWidth, dstHeight, leftPad, format, data);
  }

  /*
   * We are going to realize the operation
   * on the real display. Let's check if
   * the link is down.
   */
   
  if (NXQueryDisplayError(nxagentDisplay) == 1)
  {
    goto nxagentPutImageEnd;
  }

  /*
   * Mark the region as corrupted and skip the operation
   * if we went out of bandwidth. The drawable will be
   * synchronized at later time. Don't do that if the
   * image is likely to be a shape or a clip mask, if we
   * are here because we are actually synchronizing the
   * drawable or if the drawable's corrupted region is
   * over-age.
   */

  drawableAging = (nxagentDrawableTimestamp(pDrawable) == 0 ?
                       0 : GetTimeInMillis() - nxagentDrawableTimestamp(pDrawable));

  if (nxagentSplitTrap == 0 &&
          nxagentGetLazyLevel() > 0 &&
              drawableAging < MAX_DRAWABLE_CORRUPTED_AGING)
  {
    if (nxagentSynchronization == 0 &&
            pDrawable -> depth != 1)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentPutImage: WARNING! Prevented operation on region [%d,%d,%d,%d] "
                  "for [%s] at [%p] with link congested.\n", pRegion -> extents.x1,
                      pRegion -> extents.y1, pRegion -> extents.x2, pRegion -> extents.y2,
                          (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                              (void *) pDrawable);
      #endif
  
      nxagentMarkCorruptedRegion(pDrawable, pRegion);

      goto nxagentPutImageEnd;
    }
    else if (pDrawable -> type == DRAWABLE_PIXMAP &&
                 pDrawable -> depth != 1 &&
                     nxagentGetLazyLevel() >= 2)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentPutImage: WARNING! Prevented operation on region [%d,%d,%d,%d] "
                  "for drawable at [%p] with drawable pixmap.\n", pRegion -> extents.x1,
                      pRegion -> extents.y1, pRegion -> extents.x2, pRegion -> extents.y2,
                          (void *) pDrawable);
      #endif
  
      nxagentMarkCorruptedRegion(pDrawable, pRegion);

      goto nxagentPutImageEnd;
    }
    else if (pDrawable -> type == DRAWABLE_WINDOW &&
                 nxagentGetLazyLevel() >= 3 &&
                     nxagentGlxTrap == 0)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentPutImage: WARNING! Prevented operation on region [%d,%d,%d,%d] "
                  "for drawable at [%p] with drawable window.\n", pRegion -> extents.x1,
                      pRegion -> extents.y1, pRegion -> extents.x2, pRegion -> extents.y2,
                          (void *) pDrawable);
      #endif
  
      nxagentMarkCorruptedRegion(pDrawable, pRegion);

      goto nxagentPutImageEnd;
    }
  }
  #ifdef TEST
  else
  {
    fprintf(stderr, "nxagentPutImage: Operation on drawable [%p] not skipped with nxagentSpliTrap [%d] "
                "Aging [%lu].\n", (void *) pDrawable, nxagentSplitTrap, drawableAging);
  }
  #endif

  /*
   * Check whether we need to enclose the
   * image in a split sequence.
   */
/*
FIXME: Should we disable the split with link LAN?

  split = (nxagentOption(Streaming) == 1 &&
               nxagentOption(LinkType) != LINK_TYPE_NONE &&
                   nxagentOption(LinkType) != LINK_TYPE_LAN
*/
/*
FIXME: Temporarily turn off the streaming of images.
*/
  #ifdef WARNING

  if (nxagentOption(Streaming) == 1)
  {
    fprintf(stderr, "nxagentPutImage: WARNING! Option streaming is set to [%d].\n",
                 nxagentOption(Streaming));

    nxagentChangeOption(Streaming, 0);
  }

  #endif

  split = (nxagentOption(Streaming) == 1 &&
               nxagentOption(LinkType) != LINK_TYPE_NONE
/*
FIXME: Do we stream the images from GLX or Xv? If we do that,
       we should also write on the frame buffer, including the
       images put on windows, to be able to reconstruct the
       region that is out of sync. Surely we should not try to
       cache the GLX and Xv images in memory or save them in
       the image cache on disk.
*/
/*
FIXME: Temporarily stream the GLX data.

                           && nxagentGlxTrap == 0
                               && nxagentXvTrap == 0
*/
);

  /*
   * Always try to split the images during the
   * reconnection. Do that only if the link is
   * in congestion state during normal opera-
   * tion. Never split images whose depth is
   * less than 15.
   */
/*
FIXME: Temporarily always stream the images.
*/
/*
  if (split == 1 && ((nxagentCongestion == 0 &&
          nxagentReconnectTrap == 0) || nxagentSplitTrap == 1 ||
              depth < 15))
*/
  if (split == 1 && (nxagentSplitTrap == 1 || depth < 15))
  {
    #ifdef TEST

    if (nxagentCongestion == 1 ||
            nxagentSplitTrap == 1 ||
                nxagentReconnectTrap == 1)
    {
      fprintf(stderr, "nxagentPutImage: Not splitting with congestion [%d] reconnection [%d] "
                  "trap [%d] depth [%d].\n", nxagentCongestion, nxagentSplitTrap,
                      nxagentReconnectTrap, depth);
    }

    #endif

    split = 0;
  }
  #ifdef TEST
  else if (split == 1)
  {
    fprintf(stderr, "nxagentPutImage: Splitting with congestion [%d] reconnection [%d] "
                "trap [%d] depth [%d].\n", nxagentCongestion, nxagentSplitTrap,
                    nxagentReconnectTrap, depth);
  }
  #endif

  #ifdef TEST

  if (split == 1)
  {
    fprintf(stderr, "nxagentPutImage: Splitting the image with size [%d] "
                "link [%d] GLX [%d] Xv [%d].\n", length, nxagentOption(LinkType),
                    nxagentGlxTrap, nxagentXvTrap);
  }
  else if (nxagentOption(LinkType) != LINK_TYPE_NONE)
  {
    fprintf(stderr, "nxagentPutImage: Not splitting the image with size [%d] "
                "link [%d] GLX [%d] Xv [%d].\n", length, nxagentOption(LinkType),
                    nxagentGlxTrap, nxagentXvTrap);
  }

  #endif

  /*
   * If the image was originated by a GLX
   * or Xvideo request, temporarily disable
   * the use of the cache.
   */

  if (nxagentOption(LinkType) != LINK_TYPE_NONE &&
          (nxagentGlxTrap == 1 || nxagentXvTrap == 1))
  {
    #ifdef TEST
    fprintf(stderr, "nxagentPutImage: Disabling the use of the cache with GLX or Xvideo.\n");
    #endif

    NXSetCacheParameters(nxagentDisplay, 0, 1, 0, 0);

    cache = 0;
  }

  /*
   * Enclose the next messages in a split
   * sequence. The proxy will tell us if
   * the split took place.
   */

  if (split == 1)
  {
    /*
     * If the drawable is already being split,
     * expand the region. Currently drawables
     * can't have more than a single split
     * region.
     */

    if (nxagentSplitResource(pDrawable) != NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentPutImage: WARNING! Expanding the region with drawable at [%p] streaming.\n",
                  (void *) pDrawable);
      #endif
/*
FIXME: Should probably intersect the region with
       the region being split to also invalidate
       the commits.
*/
      nxagentMarkCorruptedRegion(pDrawable, pRegion);

      goto nxagentPutImageEnd;
    }
    else
    {
      /*
       * Assign a new resource to the drawable.
       * Will also assign the GC to use for the
       * operation.
       */

      resource = nxagentCreateSplit(pDrawable, &pGC);

      #ifdef TEST
      fprintf(stderr, "nxagentPutImage: Resource [%d] assigned to drawable at [%p].\n",
                  resource, (void *) pDrawable);
      #endif
    }

    NXStartSplit(nxagentDisplay, resource, NXSplitModeDefault);
  }

  nxagentRealizeImage(pDrawable, pGC, depth, dstX, dstY,
                          dstWidth, dstHeight, leftPad, format, data);

  if (split == 1)
  {
    NXEndSplit(nxagentDisplay, resource);

    /*
     * Now we need to check if all the messages went
     * straight through the output stream or any of
     * them required a split. If no split will take
     * place, we will remove the association with the
     * drawable and release the resource at the time
     * we will handle the no-split event.
     */

    split = nxagentWaitSplitEvent(resource);

    if (split == 1)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentPutImage: Marking corrupted region [%d,%d,%d,%d] for drawable at [%p].\n",
                  pRegion -> extents.x1, pRegion -> extents.y1, pRegion -> extents.x2,
                      pRegion -> extents.y2, (void *) pDrawable);
      #endif

      /*
       * Marking the corrupted region we will check
       * if the region intersects the split region,
       * therefore the split region must be added
       * later.
       */

      nxagentMarkCorruptedRegion(pDrawable, pRegion);

      /*
       * Assign the region to the drawable.
       */

      nxagentRegionSplit(pDrawable, pRegion);

      pRegion = NullRegion;
    }
  }

  /*
   * The split value could be changed by a
   * no-split event in the block above, so
   * here we have to check the value again.
   */

  if (split == 0)
  {
    if (nxagentDrawableStatus(pDrawable) == NotSynchronized)
    {
      /*
       * We just covered the drawable with
       * a solid image. We can consider the
       * overlapping region as synchronized.
       */

      #ifdef TEST
      fprintf(stderr, "nxagentPutImage: Marking synchronized region [%d,%d,%d,%d] for drawable at [%p].\n",
                  pRegion -> extents.x1, pRegion -> extents.y1, pRegion -> extents.x2,
                      pRegion -> extents.y2, (void *) pDrawable);
      #endif

      nxagentUnmarkCorruptedRegion(pDrawable, pRegion);
    }
  }

  /*
   * Forced synchronization based on the drawable aging.
   */

  if (drawableAging > MAX_DRAWABLE_CORRUPTED_AGING &&
          nxagentSplitTrap == 0)
  {
    nxagentSynchronizeRegion(pDrawable, NullRegion);
  }

nxagentPutImageEnd:

  /*
   * Check if we disabled caching.
   */

  if (cache == 0)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentPutImage: Reenabling the use of the cache.\n");
    #endif

    NXSetCacheParameters(nxagentDisplay, 1, 1, 1, 1);
  }

  if (pRegion != NullRegion)
  {
    nxagentFreeRegion(pDrawable, pRegion);
  }
}

void nxagentRealizeImage(DrawablePtr pDrawable, GCPtr pGC, int depth,
                             int x, int y, int w, int h, int leftPad,
                                 int format, char *data)
{
  int length;

  int bytesPerLine;
  int numSubImages;
  int totalHeight;

  const int subSize = (MAX_REQUEST_SIZE << 2) - sizeof(xPutImageReq);

  Visual *pVisual = NULL;

  RegionPtr clipRegion = NullRegion;

  /*
   * Get the visual according to the
   * drawable and depth.
   */

  pVisual = nxagentImageVisual(pDrawable, depth);

  /*
   * Get bytes per line according to format.
   */

  bytesPerLine = nxagentImagePad(w, format, leftPad, depth);

  totalHeight = h;

  length = bytesPerLine * h;

  h = (subSize < length ? subSize : length) / bytesPerLine;

  numSubImages = totalHeight / h + 1;

  while (numSubImages > 0)
  {
/*
FIXME: Testing speed with the lowest quality.
*/
/*
    nxagentPackLossless = PACK_RGB_16M_COLORS;
    nxagentPackMethod   = PACK_JPEG_16M_COLORS;
    nxagentPackQuality  = 1;
*/

    if (pDrawable -> type == DRAWABLE_WINDOW)
    {
      clipRegion = nxagentCreateRegion(pDrawable, pGC, x, y, w, h);
    }

    if (clipRegion == NullRegion || REGION_NIL(clipRegion) == 0)
    {
      nxagentPutSubImage(pDrawable, pGC, depth, x, y, w, h,
                             leftPad, format, data, pVisual);
    }
    #ifdef TEST
    else
    {
      fprintf(stderr, "nxagentRealizeImage: Skipping the put sub image with geometry "
                  "[%d,%d,%d,%d] on hidden window at [%p].\n", x, y, w, h, (void *) pDrawable);
    }
    #endif

    if (clipRegion != NullRegion)
    {
      nxagentFreeRegion(pDrawable, clipRegion);
    }

    /*
     * Flush the display if the flush has
     * been requested by the proxy.
     */

    if (nxagentFlush == 1)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentRealizeImage: Flushing with [%d] bytes on demand of the NX transport.\n",
                  NXQueryDisplay(nxagentDisplay, NXDisplayFlushable));
      #endif

      nxagentFlush = 0;

      NXFlushDisplay(nxagentDisplay, NXFlushLink);
    }

    y += h;

    data += h * bytesPerLine;

    if (--numSubImages == 1)
    {
      h = totalHeight % h;

      if (h == 0)
      {
        break;
      }
    }
  }

}

void nxagentPutSubImage(DrawablePtr pDrawable, GCPtr pGC, int depth,
                            int x, int y, int w, int h, int leftPad, int format,
                                char *data, Visual *pVisual)
{
  NXPackedImage *packedImage    = NULL;
  XImage        *plainImage     = NULL;
  unsigned char *packedChecksum = NULL;

  unsigned int packMethod  = nxagentPackMethod;
  unsigned int packQuality = nxagentPackQuality;

  ClientPtr client;

  int lossless = 0;
  int clean    = 0;
  int pack     = 0;

  /*
   * XCreateImage is the place where the leftPad should be passed.
   * The image data is received from our client unmodified. In
   * theory what we would need to do is just creating an appropri-
   * ate XImage structure based on the incoming data and let Xlib
   * do the rest. Probably we don't have to pass leftPad again in
   * the src_x of XPutImage otherwise the src_x would make Xlib
   * to take into account the xoffset field twice. Unfortunately
   * passing the leftPad doesn't work.
   *
   * plainImage = XCreateImage(nxagentDisplay, pVisual,
   *                           depth, format, leftPad, (char *) data,
   *                           w, h, BitmapPad(nxagentDisplay),
   *                           nxagentImagePad(w, format, leftPad, depth));
   */

  if ((plainImage = XCreateImage(nxagentDisplay, pVisual,
                                 depth, format, leftPad, (char *) data,
                                 w, h, BitmapPad(nxagentDisplay),
                                 nxagentImagePad(w, format, leftPad, depth))) == NULL)

  {
    #ifdef WARNING
    fprintf(stderr, "nxagentPutSubImage: WARNING! Failed to create image.\n");
    #endif

    goto nxagentPutSubImageEnd;
  }

  #ifdef DEBUG
  fprintf(stderr, "nxagentPutSubImage: Handling image with geometry [%d,%d] depth [%d].\n",
              w, h, depth);

  fprintf(stderr, "nxagentPutSubImage: Default pack method is [%d] quality is [%d].\n",
              packMethod, packQuality);
  #endif

/*
FIXME: Should use an unpack resource here.
*/
  client = requestingClient;

  if (client == NULL)
  {
    client = serverClient;
 
    #ifdef TEST
    fprintf(stderr, "nxagentPutSubImage: WARNING! Using the server client with index [%d].\n",
                client -> index);
    #endif
  }

  #ifdef TEST
  fprintf(stderr, "nxagentPutSubImage: Server order is [%d] client order is [%d].\n",
              nxagentServerOrder(), nxagentClientOrder(client));
  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentPutSubImage: Display image order is [%d] bitmap order is [%d].\n",
              ImageByteOrder(nxagentDisplay), BitmapBitOrder(nxagentDisplay));
  #endif

  /*
   * We got the image data from the X client or
   * from the frame-buffer with our own endianess.
   * Byte swap the image data if the display has
   * a different endianess than our own.
   */

  if (nxagentImageNormalize(plainImage) != 0)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentPutSubImage: WARNING! Reformatted the image with remote order [%d/%d].\n",
                plainImage -> byte_order, plainImage -> bitmap_bit_order);
    #endif
  }

  #ifdef TEST
  fprintf(stderr, "nxagentPutSubImage: Image has visual with RGB color masks [%0lx][%0lx][%0lx].\n",
              pVisual -> red_mask, pVisual -> green_mask, pVisual -> blue_mask);
  #endif

  /*
   * Check if the user requested to pack the
   * image but don't pack it if we are not
   * connected to a proxy or if the depth is
   * less than 15 bpp.
   */

  pack = (nxagentOption(LinkType) != LINK_TYPE_NONE &&
              packMethod != NO_PACK && depth > 8);

  if (pack == 1)
  {
    /*
     * Force the image to be sent as a plain
     * bitmap if we don't have any lossless
     * encoder available.
     */

    if (nxagentImageLength(w, h, format, leftPad, depth) <=
            IMAGE_PACK_THRESHOLD || nxagentLosslessTrap == 1)
    {
      if (nxagentPackLossless == NO_PACK)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentPutSubImage: Disabling pack with lossless method [%d] "
                    "trap [%d] size [%d].\n", nxagentPackLossless, nxagentLosslessTrap,
                        nxagentImageLength(w, h, format, leftPad, depth));
        #endif

        pack = 0;
      }
      else
      {
        #ifdef TEST
        fprintf(stderr, "nxagentPutSubImage: Using a lossless encoding with method [%d] "
                    "trap [%d] size [%d].\n", nxagentPackLossless, nxagentLosslessTrap,
                        nxagentImageLength(w, h, format, leftPad, depth));
        #endif

        packMethod = nxagentPackLossless;

        lossless = 1;
      }
    }
    #ifdef TEST
    else
    {
      fprintf(stderr, "nxagentPutSubImage: Using the default encoding with method [%d] "
                  "trap [%d] size [%d].\n", nxagentPackLossless, nxagentLosslessTrap,
                      nxagentImageLength(w, h, format, leftPad, depth));
    }
    #endif
  }

  if (pack == 1)
  {
    /*
     * Set the geometry and alpha channel
     * to be used for the unpacked image.
     */

    if (nxagentUnpackVisualId[client -> index] != pVisual -> visualid)
    {
      nxagentUnpackVisualId[client -> index] = pVisual -> visualid;

      #ifdef DEBUG
      fprintf(stderr, "nxagentPutSubImage: Sending new geometry for client [%d].\n",
                  client -> index);
      #endif

      NXSetUnpackGeometry(nxagentDisplay, client -> index, pVisual);
    }

    /*
     * Check if the image is supposed to carry
     * the alpha data in the fourth byte and,
     * if so, send the alpha channel using the
     * specific encoding.
     */

    if (plainImage -> depth == 32)
    {
      nxagentSetUnpackAlpha(pDrawable, plainImage, client);
    }

    /*
     * If the image doesn't come from the XVideo or the
     * GLX extension try to locate it in the cache. The
     * case of the lossless trap is also special, as we
     * want to eventually encode the image again using
     * a lossless compression.
     */
/*
FIXME: Should try to locate the image anyway, if the lossless
       trap is set, and if the image was encoded by a lossy
       compressor, roll back the changes and encode the image
       again using the preferred method.
*/
    if (nxagentGlxTrap == 0 && nxagentXvTrap == 0 &&
            nxagentLosslessTrap == 0 && NXImageCacheSize > 0)
    {
      /*
       * Be sure that the padding bits are
       * cleaned before calculating the MD5
       * checksum.
       */
/*
FIXME: There should be a callback registered by the agent that
       provides a statistics report, in text format, telling
       for example how many images were searched in the cache,
       how many were found, how many drawables are to be synch-
       ronized, etc. This statistics report would be included
       by the proxy in its stat output.
*/
      clean = 1;

      NXCleanImage(plainImage);

      /*
       * Will return a pointer to the image and checksum
       * taken from the cache, if found. If the image is
       * not found, the function returns a null image and
       * a pointer to the calculated checksum. It is up
       * to the application to free the memory. We will
       * use the checksum to add the image in the cache.
       */

      packedImage = NXCacheFindImage(plainImage, &packMethod, &packedChecksum);

      nxagentImageStatistics.partialLookups++;
      nxagentImageStatistics.totalLookups++;

      if (packedImage != NULL)
      {
        #ifdef DEBUG
        fprintf(stderr, "nxagentPutSubImage: Matched image of geometry [%d,%d] data size [%d] in cache.\n",
                    plainImage -> width, plainImage -> height, packedImage -> xoffset);
        #endif

        NXPutPackedImage(nxagentDisplay, client -> index, nxagentDrawable(pDrawable),
                             nxagentGC(pGC), packedImage, packMethod, depth,
                                 0, 0, x, y, w, h);

        nxagentImageStatistics.partialFound++;
        nxagentImageStatistics.totalFound++;

        packedChecksum = NULL;
        packedImage    = NULL;

        goto nxagentPutSubImageEnd;
      }
      #ifdef DEBUG
      else
      {
        fprintf(stderr, "nxagentPutSubImage: WARNING! Missed image of geometry [%d,%d] in cache.\n",
                    plainImage -> width, plainImage -> height);
      }
      #endif
    }

/*
FIXME: We should select the best pack method here, by
       selecting either a lossy or lossless encoding,
       unless the lossless flag was set for the image.

    if (lossless == 0)
    {
      ...
    }
*/

    /*
     * Encode the image using the selected
     * pack method.
     */

    if (packMethod == PACK_RLE_16M_COLORS ||
            packMethod == PACK_RGB_16M_COLORS)
    {
      /*
       * Cleanup the image if we didn't do that yet.
       * We assume that the JPEG and PNG compression
       * methods will actually ignore the padding
       * bytes. In other words, bitmap images prod-
       * ucing the same visual output should produce
       * compressed images that are bitwise the same
       * regardless the padding bits.
       */

      if (clean == 0)
      {
        clean = 1;

        NXCleanImage(plainImage);
      }

      if (packMethod == PACK_RLE_16M_COLORS)
      {
        packedImage = NXEncodeRle(plainImage, packMethod, packQuality);
      }
      else
      {
        packedImage = NXEncodeRgb(plainImage, packMethod, packQuality);
      }
    }
    else if (packMethod >= PACK_JPEG_8_COLORS &&
                 packMethod <= PACK_JPEG_16M_COLORS)
    {
      packedImage = NXEncodeJpeg(plainImage, packMethod, packQuality);
    }
    else if (packMethod >= PACK_PNG_8_COLORS &&
                 packMethod <= PACK_PNG_16M_COLORS)
    {
      packedImage = NXEncodePng(plainImage, packMethod, packQuality);
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentPutSubImage: WARNING! Ignoring deprecated pack method [%d].\n",
                  packMethod);
      #endif
    }

    nxagentImageStatistics.partialEncoded++;
    nxagentImageStatistics.totalEncoded++;
  }

  /*
   * If we didn't produce a valid packed
   * image, send the image as a X bitmap.
   */

  if (packedImage != NULL)
  {
    #ifdef DEBUG
    fprintf(stderr, "nxagentPutSubImage: Using packed image with method [%d] quality [%d] "
                "geometry [%d,%d] data size [%d].\n", packMethod, packQuality,
                    plainImage -> width, plainImage -> height,
                        packedImage -> xoffset);
    #endif

    NXPutPackedImage(nxagentDisplay, client -> index, nxagentDrawable(pDrawable),
                         nxagentGC(pGC), packedImage, packMethod, depth,
                             0, 0, x, y, w, h);

    /*
     * Add the image only if we have a valid
     * checksum. This is the case only if we
     * originally tried to find the image in
     * cache.
     */
      
    if (NXImageCacheSize > 0 && packedChecksum != NULL)
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentPutSubImage: Adding image with geometry [%d,%d] data size [%d] "
                  "to the cache.\n", packedImage -> width, packedImage -> height,
                      packedImage -> xoffset);
      #endif

      NXCacheAddImage(packedImage, packMethod, packedChecksum);

      nxagentImageStatistics.partialAdded++;
      nxagentImageStatistics.totalAdded++;

      packedChecksum = NULL;
      packedImage    = NULL;
    }
  }
  else
  {
    /*
     * Clean the image to help the proxy to match
     * the checksum in its cache. Do that only if
     * the differential compression is enabled and
     * if the image is not supposed to carry the
     * alpha data in the fourth byte of the pixel.
     */
/*
FIXME: If we failed to encode the image by any of the available
       methods, for example if we couldn't allocate memory, we
       may need to ripristinate the alpha channel, that in the
       meanwhile was sent in the unpack alpha message. This can
       be done here, if the clean flag is true and we are going
       to send a plain image.
*/
    if (clean == 0)
    {
      clean = (nxagentOption(LinkType) != LINK_TYPE_NONE &&
                   nxagentOption(LinkType) != LINK_TYPE_LAN && depth != 32);

      if (clean == 1)
      {
        #ifdef DEBUG
        fprintf(stderr, "nxagentPutSubImage: Cleaning the image with link type [%d] and depth [%d].\n",
                    nxagentOption(LinkType), plainImage -> depth);
        #endif

        NXCleanImage(plainImage);
      }
      #ifdef DEBUG
      else
      {
        fprintf(stderr, "nxagentPutSubImage: Not cleaning the image with link type [%d] and depth [%d].\n",
                    nxagentOption(LinkType), plainImage -> depth);
      }
      #endif
    }

    #ifdef DEBUG
    fprintf(stderr, "nxagentPutSubImage: Calling XPutImage with geometry [%d,%d] and data size [%d].\n",
                plainImage -> width, plainImage -> height, plainImage -> bytes_per_line *
                    plainImage -> height);
    #endif

    /*
     * Passing the leftPad value in src_x doesn't work.
     *
     * XPutImage(nxagentDisplay, nxagentDrawable(pDrawable),
     *               nxagentGC(pGC), plainImage, leftPad, 0, x, y, w, h);
     */

    XPutImage(nxagentDisplay, nxagentDrawable(pDrawable),
                  nxagentGC(pGC), plainImage, 0, 0, x, y, w, h);
  }

nxagentPutSubImageEnd:

  if (packedChecksum != NULL)
  {
    Xfree(packedChecksum);
  }

  if (packedImage != NULL)
  {
    if (packedImage -> data != NULL)
    {
      Xfree(packedImage -> data);
    }

    Xfree(packedImage);
  }

  Xfree(plainImage);
}

void nxagentGetImage(DrawablePtr pDrawable, int x, int y, int w, int h,
                         unsigned int format, unsigned long planeMask, char *data)
{
  #ifdef TEST
  fprintf(stderr, "nxagentGetImage: Called with drawable at [%p] geometry [%d,%d,%d,%d].\n",
              (void *) pDrawable, x, y, w, h);

  fprintf(stderr, "nxagentGetImage: Format is [%d] plane mask is [%lx].\n",
              format, planeMask);
  #endif

  if ((pDrawable)->type == DRAWABLE_PIXMAP)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentGetImage: Going to get image from virtual pixmap at [%p].\n",
                (void *) nxagentVirtualDrawable(pDrawable));
    #endif

    fbGetImage(nxagentVirtualDrawable(pDrawable), x, y, w, h, format, planeMask, data);
  }
  else
  {
    fbGetImage(pDrawable, x, y, w, h, format, planeMask, data);
  }
}

/*
 * We have to reset the visual cache before
 * connecting to another display, so that a
 * new unpack geometry can be communicated
 * to the new proxy.
 */

void nxagentResetVisualCache()
{
  int i;

  for (i = 0; i < MAX_CONNECTIONS; i++)
  {
    nxagentUnpackVisualId[i] = None;
  }
}

void nxagentResetAlphaCache()
{
  int i;

  for (i = 0; i < MAX_CONNECTIONS; i++)
  {
    if (nxagentUnpackAlpha[i])
    {
      Xfree(nxagentUnpackAlpha[i] -> data);

      Xfree(nxagentUnpackAlpha[i]);

      nxagentUnpackAlpha[i] = NULL;
    }
  }
}
