/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "scrnintstr.h"
#include "miscstruct.h"
#include "pixmapstr.h"
#include "dixstruct.h"
#include "regionstr.h"
#include "../../include/gc.h"
#include "servermd.h"
#include "mi.h"

#include "../../fb/fb.h"

#include "Agent.h"
#include "Display.h"
#include "Screen.h"
#include "Pixmap.h"
#include "Trap.h"
#include "GC.h"
#include "GCOps.h"
#include "Image.h"
#include "Split.h"
#include "Drawable.h"
#include "Visual.h"
#include "Client.h"
#include "Events.h"
#include "Holder.h"

#include "NXlib.h"
#include "NXpack.h"

RESTYPE  RT_NX_PIXMAP;

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG
#undef  DUMP

#ifdef TEST
#include "Font.h"
#endif

int nxagentPixmapPrivateIndex;

int nxagentCorruptedPixmaps;
int nxagentCorruptedBackgrounds;

static GCPtr nxagentGetTempGC(PixmapPtr);
static Bool nxagentFreeAllTempGC(void);

/*
 * Force deallocation of the virtual pixmap.
 */

static Bool nxagentDestroyVirtualPixmap(PixmapPtr pPixmap);

/*
 * This serves as a tool to check the synchronization
 * between pixmaps in framebuffer and the correspondent
 * pixmaps in the real X server.
 */

#ifdef TEST
Bool nxagentCheckPixmapIntegrity(PixmapPtr pPixmap);
#endif

struct nxagentPixmapPair
{
  Pixmap pixmap;
  PixmapPtr pMap;
};

/*
 * Predicate functions used to synchronize the
 * content of the remote drawable with the data
 * stored in the virtual frame-buffer.
 */

void nxagentSynchronizePixmapPredicate(void *p0, XID x1, void *p2);
void nxagentSynchronizeWindowPredicate(void *p0, XID x1, void *p2);

PixmapPtr nxagentCreatePixmap(ScreenPtr pScreen, int width,
                                  int height, int depth)
{
  nxagentPrivPixmapPtr pPixmapPriv, pVirtualPriv;

  PixmapPtr pPixmap;
  PixmapPtr pVirtual;

  #ifdef DEBUG
  fprintf(stderr, "nxagentCreatePixmap: Creating pixmap with width [%d] "
              "height [%d] depth [%d].\n", width, height, depth);
  #endif

  /*
   * Create the pixmap structure but do
   * not allocate memory for the data.
   */

  pPixmap = AllocatePixmap(pScreen, 0);

  if (!pPixmap)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentCreatePixmap: WARNING! Failed to create pixmap with "
                "width [%d] height [%d] depth [%d].\n", width, height, depth);
    #endif

    return NullPixmap;
  }

  /*
   * Initialize the core members.
   */

  pPixmap -> drawable.type = DRAWABLE_PIXMAP;
  pPixmap -> drawable.class = 0;
  pPixmap -> drawable.pScreen = pScreen;
  pPixmap -> drawable.depth = depth;
  pPixmap -> drawable.bitsPerPixel = BitsPerPixel(depth);
  pPixmap -> drawable.id = 0;
  pPixmap -> drawable.serialNumber = NEXT_SERIAL_NUMBER;
  pPixmap -> drawable.x = 0;
  pPixmap -> drawable.y = 0;
  pPixmap -> drawable.width = width;
  pPixmap -> drawable.height = height;
  pPixmap -> devKind = 0;
  pPixmap -> refcnt = 1;
  pPixmap -> devPrivate.ptr = NULL;

  /*
   * Initialize the privates of the real picture.
   */

  pPixmapPriv = nxagentPixmapPriv(pPixmap);

  pPixmapPriv -> isVirtual = False;
  pPixmapPriv -> isShared = nxagentShmPixmapTrap;

  pPixmapPriv -> corruptedRegion = REGION_CREATE(pPixmap -> drawable.pScreen, (BoxRec *) NULL, 1);

  pPixmapPriv -> corruptedBackground = 0;
  pPixmapPriv -> corruptedTimestamp = 0;

  pPixmapPriv -> splitResource = NULL;

  pPixmapPriv -> isBackingPixmap = 0;

  /*
   * Create the pixmap based on the default
   * windows. The proxy knows this and uses
   * this information to optimize encode the
   * create pixmap message by including the
   * id of the drawable in the checksum.
   */

  if (width != 0 && height != 0 && nxagentGCTrap == 0)
  {
    pPixmapPriv -> id = XCreatePixmap(nxagentDisplay,
                                      nxagentDefaultWindows[pScreen -> myNum],
                                      width, height, depth);
  }
  else
  {
    pPixmapPriv -> id = 0;

    #ifdef TEST
    fprintf(stderr, "nxagentCreatePixmap: Skipping the creation of pixmap at [%p] on real "
                "X server with nxagentGCTrap [%d].\n", (void *) pPixmap, nxagentGCTrap);
    #endif
  }

  pPixmapPriv -> mid = FakeClientID(serverClient -> index);

  AddResource(pPixmapPriv -> mid, RT_NX_PIXMAP, pPixmap);

  pPixmapPriv -> pRealPixmap = pPixmap;
  pPixmapPriv -> pVirtualPixmap = NULL;
  pPixmapPriv -> pPicture = NULL;

  /*
   * Create the pixmap in the virtual framebuffer.
   */

  pVirtual = fbCreatePixmap(pScreen, width, height, depth);

  if (pVirtual == NULL)
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentCreatePixmap: PANIC! Failed to create virtual pixmap with "
                "width [%d] height [%d] depth [%d].\n", width, height, depth);
    #endif

    nxagentDestroyPixmap(pPixmap);

    return NullPixmap;
  }

  #ifdef TEST
  fprintf(stderr,"nxagentCreatePixmap: Allocated memory for the Virtual %sPixmap %p of real Pixmap %p (%dx%d)\n",
              nxagentShmPixmapTrap ? "Shm " : "", (void *) pVirtual, (void *) pPixmap, width, height);
  #endif

  pPixmapPriv -> pVirtualPixmap = pVirtual;

  /*
   * Initialize the privates of the virtual picture. We
   * could avoid to use a flag and just check the pointer
   * to the virtual pixmap that, if the pixmap is actually
   * virtual, will be NULL. Unfortunately the flag can be
   * changed in nxagentValidateGC(). That code should be
   * removed in future.
   */

  pVirtualPriv = nxagentPixmapPriv(pVirtual);

  pVirtualPriv -> isVirtual = True;
  pVirtualPriv -> isShared = nxagentShmPixmapTrap;

  pVirtualPriv -> corruptedRegion = REGION_CREATE(pVirtual -> drawable.pScreen, (BoxRec *) NULL, 1);

  pVirtualPriv -> corruptedBackground = 0;
  pVirtualPriv -> corruptedTimestamp = 0;

  pVirtualPriv -> splitResource = NULL;

  /*
   * We might distinguish real and virtual pixmaps by
   * checking the pointers to pVirtualPixmap. We should
   * also remove the copy of id and use the one of the
   * real pixmap.
   */
   
  pVirtualPriv -> id = pPixmapPriv -> id;
  pVirtualPriv -> mid = 0;

  /*
   * Storing a pointer back to the real pixmap is
   * silly. Unfortunately this is the way it has
   * been originally implemented. See also the
   * comment in destroy of the pixmap.
   */

  pVirtualPriv -> pRealPixmap = pPixmap;
  pVirtualPriv -> pVirtualPixmap = NULL;
  pVirtualPriv -> pPicture = NULL;

  /*
   * Check that the virtual pixmap is created with
   * the appropriate bits-per-plane, otherwise free
   * everything and return.
   */

  if (pVirtual -> drawable.bitsPerPixel == 0)
  {
    #ifdef WARNING

    fprintf(stderr, "nxagentCreatePixmap: WARNING! Virtual pixmap at [%p] has invalid "
                "bits per pixel.\n", (void *) pVirtual);

    fprintf(stderr, "nxagentCreatePixmap: WARNING! Real pixmap created with width [%d] "
                "height [%d] depth [%d] bits per pixel [%d].\n", pPixmap -> drawable.width,
                    pPixmap -> drawable.height = height, pPixmap -> drawable.depth,
                        pPixmap -> drawable.bitsPerPixel);

    #endif

    if (!nxagentRenderTrap)
    {
      #ifdef WARNING
      fprintf(stderr, "Warning: Disabling render extension due to missing pixmap format.\n");
      #endif

      nxagentRenderTrap = 1;
    }

    nxagentDestroyPixmap(pPixmap);

    return NullPixmap;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentCreatePixmap: Created pixmap at [%p] virtual at [%p] with width [%d] "
              "height [%d] depth [%d].\n", (void *) pPixmap, (void *) pVirtual,
                  width, height, depth);
  #endif

  return pPixmap;
}

Bool nxagentDestroyPixmap(PixmapPtr pPixmap)
{
  PixmapPtr pVirtual;

  nxagentPrivPixmapPtr pPixmapPriv;

  if (!pPixmap)
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentDestroyPixmap: PANIC! Invalid attempt to destroy "
                "a null pixmap pointer.\n");
    #endif

    return False;
  }

  pPixmapPriv = nxagentPixmapPriv(pPixmap);

  pVirtual = pPixmapPriv -> pVirtualPixmap;

  #ifdef TEST
  fprintf(stderr, "nxagentDestroyPixmap: Destroying pixmap at [%p] with virtual at [%p].\n",
              (void *) pPixmap, (void *) pVirtual);
  #endif

  if (pPixmapPriv -> isVirtual)
  {
    int refcnt;

    /*
     * For some pixmaps we receive the destroy only for the
     * virtual. Infact to draw in the framebuffer we can use
     * the virtual pixmap instead of the pointer to the real
     * one. As the virtual pixmap can collect references, we
     * must transfer those references to the real pixmap so
     * we can continue as the destroy had been requested for
     * it.
     */

    pVirtual = pPixmap;
    pPixmap  = pPixmapPriv -> pRealPixmap;

    pPixmapPriv = nxagentPixmapPriv(pPixmap);

    /*
     * Move the references accumulated by the virtual
     * pixmap into the references of the real one.
     */

    refcnt = pVirtual -> refcnt - 1;

    #ifdef TEST
    fprintf(stderr, "nxagentDestroyPixmap: Adding [%d] references to pixmap at [%p].\n",
                refcnt, (void *) pPixmap);
    #endif

    pPixmap -> refcnt += refcnt;

    pVirtual -> refcnt -= refcnt;
  }

  --pPixmap -> refcnt;

  #ifdef TEST

  fprintf(stderr, "nxagentDestroyPixmap: Pixmap has now [%d] references with virtual pixmap [%d].\n",
              pPixmap -> refcnt, pVirtual -> refcnt);

  if (pVirtual != NULL && pVirtual -> refcnt != 1)
  {
    fprintf(stderr, "nxagentDestroyPixmap: PANIC! Virtual pixmap has [%d] references.\n",
                pVirtual -> refcnt);
  }

  #endif

  if (pPixmap -> refcnt > 0)
  {
    return True;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentDestroyPixmap: Managing to destroy the pixmap at [%p]\n",
              (void *) pPixmap);
  #endif

  nxagentRemoveItemBSPixmapList(nxagentPixmap(pPixmap));

  nxagentDestroyVirtualPixmap(pPixmap);

  if (pPixmapPriv -> corruptedRegion != NullRegion)
  {
    if (pPixmapPriv -> isVirtual == 0 &&
            nxagentDrawableStatus((DrawablePtr) pPixmap) == NotSynchronized)
    {
      nxagentCorruptedPixmaps--;
    }

    REGION_DESTROY(pPixmap -> drawable.pScreen, pPixmapPriv -> corruptedRegion);

    pPixmapPriv -> corruptedRegion = NullRegion;
  }

  if (pPixmapPriv -> splitResource != NULL)
  {
    nxagentReleaseSplit((DrawablePtr) pPixmap);
  }

  if (pPixmapPriv -> corruptedBackground == 1)
  {
/*
FIXME: Should we force a synchronization and change the
       attributes of the window if the background pixmap
       is about to be destroyed?
*/
    #ifdef TEST
    fprintf(stderr, "nxagentDestroyPixmap: WARNING! Destroying the background pixmap at [%p].\n",
                (void *) pPixmap);
    #endif

    nxagentCorruptedBackgrounds--;
  }

  /*
   * A pixmap with width and height set to 0 is
   * created at the beginning. To this pixmap is
   * not assigned an id. This is likely a scratch
   * pixmap used by the X server.
   */

  if (pPixmapPriv -> id)
  {
    XFreePixmap(nxagentDisplay, pPixmapPriv -> id);
  }

  if (pPixmapPriv -> mid)
  {
    FreeResource(pPixmapPriv -> mid, RT_NONE);
  }

  xfree(pPixmap);

  return True;
}

Bool nxagentDestroyVirtualPixmap(PixmapPtr pPixmap)
{
  PixmapPtr pVirtual;
  nxagentPrivPixmapPtr pVirtualPriv;

  pVirtual = nxagentPixmapPriv(pPixmap) -> pVirtualPixmap;

  /*
   * Force the routine to get rid of the virtual
   * pixmap.
   */

  if (pVirtual != NULL)
  {
    pVirtual -> refcnt = 1;

    pVirtualPriv = nxagentPixmapPriv(pVirtual);

    if (pVirtualPriv -> corruptedRegion != NullRegion)
    {
      REGION_DESTROY(pVirtual -> drawable.pScreen, pVirtualPriv -> corruptedRegion);

      pVirtualPriv -> corruptedRegion = NullRegion;
    }

    fbDestroyPixmap(pVirtual);
  }

  return True;
}

RegionPtr nxagentPixmapToRegion(PixmapPtr pPixmap)
{
  #ifdef TEST
  fprintf(stderr, "PixmapToRegion: Pixmap = [%lx] nxagentVirtualPixmap = [%lx]\n",
              pPixmap, nxagentVirtualPixmap(pPixmap));
  #endif

  return fbPixmapToRegion(nxagentVirtualPixmap(pPixmap));
}

Bool nxagentModifyPixmapHeader(PixmapPtr pPixmap, int width, int height, int depth,
                                   int bitsPerPixel, int devKind, pointer pPixData)
{
  PixmapPtr pVirtualPixmap;

  /*
   * See miModifyPixmapHeader() in miscrinit.c. This
   * function is used to recycle the scratch pixmap
   * for this screen. We let it refer to the virtual
   * pixmap.
   */

  if (!pPixmap)
  {
    return False;
  }

  if (nxagentPixmapIsVirtual(pPixmap))
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentModifyPixmapHeader: PANIC! Pixmap at [%p] is virtual.\n",
                (void *) pPixmap);
    #endif

    FatalError("nxagentModifyPixmapHeader: PANIC! Pixmap is virtual.");
  }

  pVirtualPixmap = nxagentVirtualPixmap(pPixmap);

  #ifdef TEST
  fprintf(stderr, "nxagentModifyPixmapHeader: Pixmap has width [%d] height [%d] depth [%d] "
              "bits-per-pixel [%d] devKind [%d] pPixData [%p].\n", pPixmap->drawable.width,
                  pPixmap->drawable.height, pPixmap->drawable.depth, pPixmap->drawable.bitsPerPixel,
                      pPixmap->devKind, (void *) pPixmap->devPrivate.ptr);

  fprintf(stderr, "nxagentModifyPixmapHeader: New parameters are width [%d] height [%d] depth [%d] "
              "bits-per-pixel [%d] devKind [%d] pPixData [%p].\n", width, height, depth,
                  bitsPerPixel, devKind, (void *) pPixData);
  #endif

  if ((width > 0) && (height > 0) && (depth > 0) &&
          (bitsPerPixel > 0) && (devKind > 0) && pPixData)
  {
    pPixmap->drawable.depth = depth;
    pPixmap->drawable.bitsPerPixel = bitsPerPixel;
    pPixmap->drawable.id = 0;
    pPixmap->drawable.serialNumber = NEXT_SERIAL_NUMBER;
    pPixmap->drawable.x = 0;
    pPixmap->drawable.y = 0;
    pPixmap->drawable.width = width;
    pPixmap->drawable.height = height;
    pPixmap->devKind = devKind;
    pPixmap->refcnt = 1;
    pPixmap->devPrivate.ptr = pPixData;

    pVirtualPixmap->drawable.depth = depth;
    pVirtualPixmap->drawable.bitsPerPixel = bitsPerPixel;
    pVirtualPixmap->drawable.id = 0;
    pVirtualPixmap->drawable.serialNumber = NEXT_SERIAL_NUMBER;
    pVirtualPixmap->drawable.x = 0;
    pVirtualPixmap->drawable.y = 0;
    pVirtualPixmap->drawable.width = width;
    pVirtualPixmap->drawable.height = height;
    pVirtualPixmap->devKind = devKind;
    pVirtualPixmap->refcnt = 1;
    pVirtualPixmap->devPrivate.ptr = pPixData;
  }
  else
  {
    if (width > 0)
        pPixmap->drawable.width = width;

    if (height > 0)
        pPixmap->drawable.height = height;

    if (depth > 0)
        pPixmap->drawable.depth = depth;

    if (bitsPerPixel > 0)
        pPixmap->drawable.bitsPerPixel = bitsPerPixel;
    else if ((bitsPerPixel < 0) && (depth > 0))
        pPixmap->drawable.bitsPerPixel = BitsPerPixel(depth);

    if (devKind > 0)
        pPixmap->devKind = devKind;
    else if ((devKind < 0) && ((width > 0) || (depth > 0)))
        pPixmap->devKind = PixmapBytePad(pPixmap->drawable.width,
            pPixmap->drawable.depth);

    if (pPixData)
       pPixmap->devPrivate.ptr = pPixData; 

     /*
      * XXX This was the previous assignment:
      *
      * pVirtualPixmap->devPrivate.ptr = pPixData;
      */

    if (width > 0)
        pVirtualPixmap->drawable.width = width;

    if (height > 0)
        pVirtualPixmap->drawable.height = height;

    if (depth > 0)
        pVirtualPixmap->drawable.depth = depth;

    if (bitsPerPixel > 0)
        pVirtualPixmap->drawable.bitsPerPixel = bitsPerPixel;
    else if ((bitsPerPixel < 0) && (depth > 0))
        pVirtualPixmap->drawable.bitsPerPixel = BitsPerPixel(depth);

    if (devKind > 0)
        pVirtualPixmap->devKind = devKind;
    else if ((devKind < 0) && ((width > 0) || (depth > 0)))
        pVirtualPixmap->devKind = PixmapBytePad(pVirtualPixmap->drawable.width,
            pVirtualPixmap->drawable.depth);

    if (pPixData)
        pVirtualPixmap->devPrivate.ptr = pPixData;

    #ifdef PANIC

    if (pPixmap->drawable.x != 0 || pPixmap->drawable.y != 0)
    {
      fprintf(stderr, "nxagentModifyPixmapHeader: PANIC! Pixmap at [%p] has x [%d] and y [%d].\n",
                  (void *) pPixmap, pPixmap->drawable.x, pPixmap->drawable.y);

      FatalError("nxagentModifyPixmapHeader: PANIC! Pixmap has x or y greater than zero.");
    }

    #endif
  }

  return True;
}

static void nxagentPixmapMatchID(void *p0, XID x1, void *p2)
{
  PixmapPtr pPixmap = (PixmapPtr)p0;
  struct nxagentPixmapPair *pPair = p2;

  if ((pPair -> pMap == NULL) && (nxagentPixmap(pPixmap) == pPair -> pixmap))
  {
    pPair -> pMap = pPixmap;
  }
}

PixmapPtr nxagentPixmapPtr(Pixmap pixmap)
{
  int i;
  struct nxagentPixmapPair pair;

  if (pixmap == None)
  {
    return NULL;
  }

  pair.pixmap = pixmap;
  pair.pMap = NULL;

  FindClientResourcesByType(clients[serverClient -> index], RT_NX_PIXMAP,
                                nxagentPixmapMatchID, &pair);

  for (i = 0; (pair.pMap == NULL) && (i < MAXCLIENTS); i++)
  {
    if (clients[i])
    {
      FindClientResourcesByType(clients[i], RT_PIXMAP,
                                    nxagentPixmapMatchID, &pair);
    }
  }

  #ifdef WARNING

  if (pair.pMap == NULL)
  {
    fprintf(stderr, "nxagentFindPixmap: WARNING! Failed to find "
	    "remote pixmap [%ld].\n", (long int) pair.pixmap);
  }
  else if (nxagentDrawableStatus((DrawablePtr) pair.pMap) == NotSynchronized)
  {
    fprintf(stderr, "WARNING! Rootless icon at [%p] [%d,%d] is not synchronized.\n",
                (void *) pair.pMap, pair.pMap -> drawable.width,
                    pair.pMap -> drawable.height);
  }

  #endif

  return pair.pMap;
}

/*
 * Reconnection stuff.
 */

int nxagentDestroyNewPixmapResourceType(pointer p, XID id)
{
  /*
   * Address of the destructor is set in Init.c.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentDestroyNewPixmapResourceType: Destroying mirror id [%ld] for pixmap at [%p].\n",
              nxagentPixmapPriv((PixmapPtr) p) -> mid, (void *) p);
  #endif

  nxagentPixmapPriv((PixmapPtr) p) -> mid = None;

  return True;
}

void nxagentDisconnectPixmap(void *p0, XID x1, void *p2)
{
  PixmapPtr pPixmap = (PixmapPtr) p0;

  Bool *pBool;

  pBool = (Bool*) p2;

  #ifdef TEST
  fprintf(stderr, "nxagentDisconnectPixmap: Called with bool [%d] and pixmap at [%p].\n",
              *pBool, (void *) pPixmap);

  fprintf(stderr, "nxagentDisconnectPixmap: Virtual pixmap is [%ld].\n",
              nxagentPixmap(pPixmap));
  #endif

  nxagentPixmap(pPixmap) = None;
}

Bool nxagentDisconnectAllPixmaps()
{
  int r = 1;
  int i;

  #ifdef TEST
  fprintf(stderr, "nxagentDisconnectAllPixmaps: Going to iterate through pixmap resources.\n");
  #endif

  /*
   * The RT_NX_PIXMAP resource type is allocated
   * only on the server client, so we don't need
   * to find it through the other clients too.
   */

  FindClientResourcesByType(clients[serverClient -> index], RT_NX_PIXMAP, nxagentDisconnectPixmap, &r);

  #ifdef WARNING

  if (r == 0)
  {
    fprintf(stderr, "nxagentDisconnectAllPixmaps: WARNING! Failed to disconnect "
                "pixmap for client [%d].\n", serverClient -> index);
  }

  #endif

  for (i = 0, r = 1; i < MAXCLIENTS; r = 1, i++)
  {
    if (clients[i])
    {
      #ifdef TEST
      fprintf(stderr, "nxagentDisconnectAllPixmaps: Going to disconnect pixmaps of client [%d].\n", i);
      #endif

      FindClientResourcesByType(clients[i], RT_PIXMAP, nxagentDisconnectPixmap, &r);

      #ifdef WARNING

      if (r == 0)
      {
        fprintf(stderr, "nxagentDisconnectAllPixmaps: WARNING! Failed to disconnect "
                    "pixmap for client [%d].\n", i);
      }

      #endif
    }
  }

  #ifdef WARNING

  if (r == 0)
  {
    fprintf(stderr, "nxagentDisconnectAllPixmaps: WARNING! Failed to disconnect "
                "pixmap for client [%d].\n", i);
  }

  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentDisconnectAllPixmaps: Pixmaps disconnection completed.\n");
  #endif

  return r;
}

void nxagentReconnectPixmap(void *p0, XID x1, void *p2)
{
  PixmapPtr pPixmap = (PixmapPtr) p0;
  Bool *pBool = (Bool*) p2;
  nxagentPrivPixmapPtr pPixmapPriv;

  if (*pBool == 0 || pPixmap == NULL ||
          NXQueryDisplayError(nxagentDisplay) == 1)
  {
    *pBool = 0;

    #ifdef TEST
    fprintf(stderr, "nxagentReconnectPixmap: Ignoring pixmap at [%p] while "
                "recovering from the error.\n", (void *) pPixmap);
    #endif

    return;
  }
  else if (pPixmap == nxagentDefaultScreen -> pScratchPixmap)
  {
    /*
     * Every time the scratch pixmap is used its
     * data is changed, so we don't need to recon-
     * nect it.
     */

    #ifdef TEST
    fprintf(stderr, "nxagentReconnectPixmap: Ignoring scratch pixmap at [%p].\n",
                (void *) pPixmap);
    #endif

    return;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectPixmap: Called with result [%d] and pixmap at [%p].\n",
              *pBool, (void *) pPixmap);

  fprintf(stderr, "nxagentReconnectPixmap: Virtual pixmap is at [%p] picture is at [%p].\n",
              (void *) nxagentPixmapPriv(pPixmap) -> pVirtualPixmap,
                  (void *) nxagentPixmapPriv(pPixmap) -> pPicture);
  #endif

  pPixmapPriv = nxagentPixmapPriv(pPixmap);

  if (pPixmap -> drawable.width && pPixmap -> drawable.height)
  {
    pPixmapPriv -> id = XCreatePixmap(nxagentDisplay,
                                      nxagentDefaultWindows[pPixmap -> drawable.pScreen -> myNum],
                                      pPixmap -> drawable.width,
                                      pPixmap -> drawable.height,
                                      pPixmap -> drawable.depth);

    nxagentPixmap(pPixmapPriv -> pVirtualPixmap) = pPixmapPriv -> id;

    #ifdef TEST
    fprintf(stderr, "nxagentReconnectPixmap: Created virtual pixmap with id [%ld] for pixmap at [%p].\n",
                nxagentPixmap(pPixmap), (void *) pPixmap);
    #endif

    if (pPixmap == (PixmapPtr) nxagentDefaultScreen -> devPrivate)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentReconnectPixmap: WARNING! Pixmap is root screen. Returning.\n");
      #endif

      return;
    }

    nxagentSplitTrap = 1;

    *pBool = nxagentSynchronizeDrawableData((DrawablePtr) pPixmap);

    nxagentSplitTrap = 0;

    if (*pBool == 0)
    {
      #ifdef PANIC
      fprintf(stderr, "nxagentReconnectPixmap: PANIC! Failed to synchronize the pixmap.\n");
      #endif
    }
  }
}

Bool nxagentReconnectAllPixmaps(void *p0)
{
  Bool result = 1;

  int i;

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectAllPixmaps: Going to recreate all pixmaps.\n");
  #endif

  /*
   * Reset the geometry and alpha information
   * used by proxy to unpack the packed images.
   */

  nxagentResetVisualCache();

  nxagentResetAlphaCache();

  /*
   * The RT_NX_PIXMAP resource type is allocated
   * only on the server client, so we don't need
   * to find it through the other clients too.
   */

  FindClientResourcesByType(clients[serverClient -> index], RT_NX_PIXMAP, nxagentReconnectPixmap, &result);

  #ifdef WARNING

  if (result == 0)
  {
    fprintf(stderr, "nxagentReconnectAllPixmaps: WARNING! Failed to reconnect "
                "pixmap for client [%d].\n", serverClient -> index);
  }

  #endif

  for (i = 0, result = 1; i < MAXCLIENTS; result = 1, i++)
  {
    if (clients[i] != NULL)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentReconnectAllPixmaps: Going to reconnect pixmaps of client [%d].\n", i);
      #endif

      /*
       * Let the pixmap be reconnected as it was an
       * image request issued by the client owning
       * the resource. The client index is used as
       * a subscript by the image routines to cache
       * the data per-client.
       */

      FindClientResourcesByType(clients[i], RT_PIXMAP, nxagentReconnectPixmap, &result);

      #ifdef WARNING

      if (result == 0)
      {
        fprintf(stderr, "nxagentReconnectAllPixmaps: WARNING! Failed to reconnect "
                    "pixmap for client [%d].\n", serverClient -> index);
      }

      #endif
    }
  }

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectAllPixmaps: Going to free all the temporary GCs.\n");
  #endif

  #ifdef WARNING

  if (nxagentFreeAllTempGC() == 0)
  {
    fprintf(stderr, "nxagentReconnectAllPixmaps: WARNING! Free of temporary GCs failed.\n");
  }

  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectAllPixmaps: Pixmaps reconnection completed.\n");
  #endif

  return result;
}

typedef struct _GCCache
{
  int depth;
  GCPtr gc;

} GCCache;

GCCache *nxagentGCCache;
int nxagentGCCacheSize = 0;

static GCPtr nxagentGetTempGC(PixmapPtr pPixmap)
{
  int i;
  int result;

  GCPtr pGC;
  GCCache *cacheGC;

  for (i = 0; i < nxagentGCCacheSize; i++)
  {
    if (pPixmap -> drawable.depth == nxagentGCCache[i].depth)
    {
      return nxagentGCCache[i].gc;
    }
  }

  cacheGC = xrealloc(nxagentGCCache, (nxagentGCCacheSize + 1) * sizeof(GCCache));

  if (cacheGC == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentGetTempGC: Can't allocate memory for the temporary GC.\n");
    #endif

    return NULL;
  }

  nxagentGCCache = cacheGC;

  pGC = CreateGC((DrawablePtr) pPixmap, 0, 0, &result);

  if (result != Success)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentGetTempGC: Failed to create temporary GC for depth [%d].\n",
                pPixmap -> drawable.depth);
    #endif

    return NULL;
  }

  nxagentGCCache[nxagentGCCacheSize].gc    = pGC;
  nxagentGCCache[nxagentGCCacheSize].depth = pPixmap -> drawable.depth;

  nxagentGCCacheSize++;

  return nxagentGetTempGC(pPixmap);
}

static Bool nxagentFreeAllTempGC(void)
{
  int i;

  int result;

  for (i = 0; i < nxagentGCCacheSize; i++)
  {
    result = FreeGC(nxagentGCCache[i].gc, (XID)0);

    if (result != Success)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentFreeAllTempGC: WARNING! Failed to free temporary GC at [%p].\n",
                  (void *) nxagentGCCache[i].gc);
      #endif
    }
  }

  xfree(nxagentGCCache);

  nxagentGCCache     = NULL;
  nxagentGCCacheSize = 0;

  return 1;
}

#ifdef TEST

static void nxagentCheckOnePixmapIntegrity(void *p0, XID x1, void *p2)
{
  PixmapPtr pPixmap = (PixmapPtr) p0;
  Bool      *pBool = (Bool*) p2;

  if (*pBool == False)
  {
    return;
  }

  if (pPixmap == nxagentDefaultScreen -> devPrivate)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentCheckOnePixmapIntegrity: pixmap %p is screen.\n",
                (void *) pPixmap);
    #endif

    return;
  }

  if (pPixmap == nxagentDefaultScreen -> PixmapPerDepth[0])
  {
    #ifdef TEST
    fprintf(stderr, "nxagentCheckOnePixmapIntegrity: pixmap %p is default stipple of screen.\n",
                (void *) pPixmap);
    #endif

    return;
  }

  *pBool = nxagentCheckPixmapIntegrity(pPixmap);
}

Bool nxagentCheckPixmapIntegrity(PixmapPtr pPixmap)
{
  Bool integrity = 1;
  XImage *image;
  char *data;
  int format;
  unsigned long plane_mask = AllPlanes;
  unsigned int width, height, length, depth;
  PixmapPtr pVirtual = nxagentVirtualPixmap(pPixmap);

  width = pPixmap -> drawable.width;
  height = pPixmap -> drawable.height;
  depth = pPixmap -> drawable.depth;
  format = (depth == 1) ? XYPixmap : ZPixmap;

  if (width && height)
  {
    length = nxagentImageLength(width, height, format, 0, depth);

    data = malloc(length);

    if (data == NULL)
    {
      FatalError("nxagentCheckPixmapIntegrity: Failed to allocate a buffer of size %d.\n", length);
    }

    image = XGetImage(nxagentDisplay, nxagentPixmap(pPixmap), 0, 0,
                          width, height, plane_mask, format);
    if (image == NULL)
    {
      FatalError("XGetImage: Failed.\n");

      return False;
    }

    #ifdef WARNING
    fprintf(stderr, "nxagentCheckPixmapIntegrity: Image from X has length [%d] and checksum [0x%s].\n",
                length, nxagentChecksum(image->data, length));
    #endif

    NXCleanImage(image);

    #ifdef WARNING
    fprintf(stderr, "nxagentCheckPixmapIntegrity: Image after clean has checksum [0x%s].\n",
                nxagentChecksum(image->data, length));
    #endif

    fbGetImage((DrawablePtr) pVirtual, 0, 0, width, height, format, plane_mask, data);

    #ifdef WARNING
    fprintf(stderr, "nxagentCheckPixmapIntegrity: Image from FB has length [%d] and checksum [0x%s].\n",
                length, nxagentChecksum(data, length));
    #endif

    if (image != NULL && memcmp(image -> data, data, length) != 0)
    {
      integrity = 0;
    }
    else
    {
      integrity = 1;

      #ifdef TEST
      fprintf(stderr, "nxagentCheckPixmapIntegrity: Pixmap at [%p] has been realized. "
                  "Now remote and frambuffer data are synchronized.\n", (void *) pPixmap);
      #endif
    }

    #ifdef WARNING

    if (integrity == 0)
    {

      int i;
      char *p, *q;

      for (i = 0, p = image -> data, q = data; i < length; i++)
      {
        if (p[i] != q[i])
        {
          fprintf(stderr, "nxagentCheckPixmapIntegrity: Byte [%d] image -> data [%d] data [%d]. "
                      "Buffers differ!\n", i, p[i], q[i]);
        }
        else
        {
          fprintf(stderr, "nxagentCheckPixmapIntegrity: Byte [%d] image -> data [%d] data [%d].\n",
                      i, p[i], q[i]);
        }
      }

      fprintf(stderr, "nxagentCheckPixmapIntegrity: Pixmap at [%p] width [%d], height [%d], has been realized "
                  "but the data buffer still differs.\n", (void *) pPixmap, width, height);

      fprintf(stderr, "nxagentCheckPixmapIntegrity: bytes_per_line [%d] byte pad [%d] format [%d].\n",
                  image -> bytes_per_line, nxagentImagePad(width, height, 0, depth), image -> format);

      FatalError("nxagentCheckPixmapIntegrity: Image is corrupted!!\n");

    }

    #endif

    if (image != NULL)
    {
      XDestroyImage(image);
    }

    if (data != NULL)
    {
      free(data);
    }
  }
  else
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentCheckPixmapIntegrity: Ignored pixmap at [%p] with geometry [%d] [%d].\n",
                (void *) pPixmap, width, height);
    #endif
  }

  return integrity;
}

Bool nxagentCheckAllPixmapIntegrity()
{
  int i;
  Bool imageIsGood = True;

  #ifdef TEST
  fprintf(stderr, "nxagentCheckAllPixmapIntegrity\n");
  #endif

  FindClientResourcesByType(clients[serverClient -> index], RT_NX_PIXMAP,
                                nxagentCheckOnePixmapIntegrity, &imageIsGood);

  for (i = 0; (i < MAXCLIENTS) && (imageIsGood); i++)
  {
    if (clients[i])
    {
      FindClientResourcesByType(clients[i], RT_PIXMAP,
                                    nxagentCheckOnePixmapIntegrity, &imageIsGood);

    }
  }

  #ifdef TEST
  fprintf(stderr, "nxagentCheckAllPixmapIntegrity: pixmaps integrity = %d.\n", imageIsGood);
  #endif

  return imageIsGood;
}

#endif

RegionPtr nxagentCreateRegion(DrawablePtr pDrawable, GCPtr pGC, int x, int y,
                                  int width, int height)
{
  RegionPtr pRegion;
  BoxRec box;

  box.x1 = x;
  box.y1 = y;
  box.x2 = x + width;
  box.y2 = y + height;

  pRegion = REGION_CREATE(pDrawable -> pScreen, &box, 1);

  /*
   * Clipping the region.
   */

  if (pDrawable -> type == DRAWABLE_PIXMAP)
  {
    BoxRec tmpBox;
    RegionRec tmpRegion;

    /*
     * The region created doesn't need to be clipped
     * if it has the pixmap dimensions.
     */

    if (x != 0 || y != 0 ||
            width != pDrawable -> width ||
                height != pDrawable -> height)
    {
      tmpBox.x1 = 0;
      tmpBox.y1 = 0;
      tmpBox.x2 = pDrawable -> width;
      tmpBox.y2 = pDrawable -> height;

      REGION_INIT(pDrawable -> pScreen, &tmpRegion, &tmpBox, 1);

      REGION_INTERSECT(pDrawable -> pScreen, pRegion, &tmpRegion, pRegion);

      REGION_UNINIT(pDrawable -> pScreen, &tmpRegion);
    }
  }
  else
  {
    REGION_TRANSLATE(pDrawable -> pScreen, pRegion,
                         pDrawable -> x + wBorderWidth((WindowPtr) pDrawable),
                             pDrawable -> y + wBorderWidth((WindowPtr) pDrawable));

    REGION_INTERSECT(pDrawable -> pScreen, pRegion, pRegion, &((WindowPtr) pDrawable) -> borderClip);

    REGION_TRANSLATE(pDrawable -> pScreen, pRegion,
                         -(pDrawable -> x - wBorderWidth((WindowPtr) pDrawable)),
                             -(pDrawable -> y - wBorderWidth((WindowPtr) pDrawable)));
  }

  #ifdef TEST
  fprintf(stderr, "nxagentCreateRegion: New region created with coordinates [%d,%d,%d,%d].\n",
              pRegion -> extents.x1, pRegion -> extents.y1,
                  pRegion -> extents.x2, pRegion -> extents.y2);
  #endif

  if (pGC != NULL && pGC -> clientClip != NULL &&
          pGC -> clientClipType == CT_REGION)
  {
    RegionRec clipRegion;

    REGION_INIT(pDrawable -> pScreen, &clipRegion, NullBox, 1);

    REGION_COPY(pDrawable -> pScreen, &clipRegion, (RegionPtr) pGC -> clientClip);

    /*
     * The clip origin is relative to the origin of
     * the destination drawable. The clip mask coor-
     * dinates are relative to the clip origin.
     */

    if (pGC -> clipOrg.x != 0 || pGC -> clipOrg.y != 0)
    {
      REGION_TRANSLATE(pDrawable -> pScreen, &clipRegion, pGC -> clipOrg.x, pGC -> clipOrg.y);
    }

    #ifdef TEST
    fprintf(stderr, "nxagentCreateRegion: Clipping region to the clip mask with coordinates [%d,%d,%d,%d].\n",
                clipRegion.extents.x1, clipRegion.extents.y1,
                    clipRegion.extents.x2, clipRegion.extents.y2);
    #endif

    REGION_INTERSECT(pDrawable -> pScreen, pRegion, pRegion, &clipRegion);

    REGION_UNINIT(pDrawable -> pScreen, &clipRegion);
  }

  return pRegion;
}

void nxagentMarkCorruptedRegion(DrawablePtr pDrawable, RegionPtr pRegion)
{
  int x;
  int y;
  int width;
  int height;

  /*
   * If the drawable was synchronized, the counter
   * reporting the number of corrupted drawables
   * must be increased. Moreover the corrupted ti-
   * mestamp must be set.
   */

  if (nxagentDrawableStatus(pDrawable) == Synchronized)
  {
    if (pDrawable -> type == DRAWABLE_PIXMAP)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentMarkCorruptedRegion: Corrupted pixmaps counter was [%d].\n",
                  nxagentCorruptedPixmaps);
      #endif

      nxagentCorruptedPixmaps++;
    }
    else
    {
      #ifdef TEST
      fprintf(stderr, "nxagentMarkCorruptedRegion: Corrupted windows counter was [%d].\n",
                  nxagentCorruptedWindows);
      #endif

      nxagentCorruptedWindows++;
    }

    nxagentSetCorruptedTimestamp(pDrawable);
  }


  if (pRegion == NullRegion)
  {
    x = 0;
    y = 0;

    width  = pDrawable -> width;
    height = pDrawable -> height;
  
    #ifdef TEST
    fprintf(stderr, "nxagentMarkCorruptedRegion: Fully invalidating %s [%p] with "
                "coordinates [%d,%d][%d,%d].\n", (pDrawable -> type == DRAWABLE_PIXMAP ?
                    "pixmap" : "window"), (void *) pDrawable, x, y, width, height);
    #endif

    pRegion = nxagentCreateRegion(pDrawable, NULL, x, y, width, height);

    nxagentValidateSplit(pDrawable, pRegion);

    REGION_UNION(pDrawable -> pScreen, nxagentCorruptedRegion(pDrawable),
                     nxagentCorruptedRegion(pDrawable), pRegion);

    nxagentFreeRegion(pDrawable, pRegion);
  }
  else
  {
    #ifdef TEST

    x = pRegion -> extents.x1;
    y = pRegion -> extents.y1;

    width = pRegion -> extents.x2 - pRegion -> extents.x1;
    height = pRegion -> extents.y2 - pRegion -> extents.y1;

    fprintf(stderr, "nxagentMarkCorruptedRegion: Partly invalidating %s [%p] with "
                "coordinates [%d,%d][%d,%d].\n", (pDrawable -> type == DRAWABLE_PIXMAP ?
                    "pixmap" : "window"), (void *) pDrawable, x, y, width, height);

    #endif

    nxagentValidateSplit(pDrawable, pRegion);

    REGION_UNION(pDrawable -> pScreen, nxagentCorruptedRegion(pDrawable),
                     nxagentCorruptedRegion(pDrawable), pRegion);
  }
}

void nxagentUnmarkCorruptedRegion(DrawablePtr pDrawable, RegionPtr pRegion)
{
  int oldStatus;

  #ifdef TEST

  int x;
  int y;
  int width;
  int height;

  #endif

  oldStatus = nxagentDrawableStatus(pDrawable);

  if (pRegion == NullRegion)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentUnmarkCorruptedRegion: Fully validating %s [%p].\n",
                (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"), 
                    (void *) pDrawable);
    #endif

    nxagentValidateSplit(pDrawable, NULL);

    REGION_EMPTY(pDrawable -> pScreen, nxagentCorruptedRegion(pDrawable));
  }
  else
  {
    #ifdef TEST

    x = pRegion -> extents.x1;
    y = pRegion -> extents.y1;

    width = pRegion -> extents.x2 - pRegion -> extents.x1;
    height = pRegion -> extents.y2 - pRegion -> extents.y1;

    fprintf(stderr, "nxagentUnmarkCorruptedRegion: Validating %s [%p] with "
                "coordinates [%d,%d,%d,%d].\n", (pDrawable -> type == DRAWABLE_PIXMAP ?
                    "pixmap" : "window"), (void *) pDrawable, x, y, width, height);

    #endif

    nxagentValidateSplit(pDrawable, pRegion);

    REGION_SUBTRACT(pDrawable -> pScreen, nxagentCorruptedRegion(pDrawable),
                        nxagentCorruptedRegion(pDrawable), pRegion);
  }

  /*
   * If the drawable becomes synchronized, the
   * counter reporting the number of corrupted
   * drawables must be decreased. Moreover the
   * corrupted timestamp must be reset.
   */

  if (oldStatus == NotSynchronized &&
          nxagentDrawableStatus(pDrawable) == Synchronized)
  {
    if (pDrawable -> type == DRAWABLE_PIXMAP)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentUnmarkCorruptedRegion: Corrupted pixmaps counter was [%d].\n",
                  nxagentCorruptedPixmaps);
      #endif

      nxagentCorruptedPixmaps--;
    }
    else
    {
      #ifdef TEST
      fprintf(stderr, "nxagentUnmarkCorruptedRegion: Corrupted windows counter was [%d].\n",
                  nxagentCorruptedWindows);
      #endif

      nxagentCorruptedWindows--;
    }

    nxagentResetCorruptedTimestamp(pDrawable);
  }
}

void nxagentMoveCorruptedRegion(WindowPtr pWin, unsigned int mask)
{
  /*
   * If a window is resized, its corrupted
   * region is moved according to the bit
   * gravity.
   */

  if (nxagentDrawableStatus((DrawablePtr) pWin) == NotSynchronized)
  {
    if (((mask & CWHeight) && nxagentWindowPriv(pWin) -> height != pWin -> drawable.height) ||
            ((mask & CWWidth) && nxagentWindowPriv(pWin) -> width != pWin -> drawable.width))
    {
      int nx, ny;

      GravityTranslate(0, 0,
                       nxagentWindowPriv(pWin) -> x - pWin -> origin.x + wBorderWidth(pWin),
                       nxagentWindowPriv(pWin) -> y - pWin -> origin.y + wBorderWidth(pWin),
                       pWin -> drawable.width - nxagentWindowPriv(pWin) -> width,
                       pWin -> drawable.height - nxagentWindowPriv(pWin) -> height,
                       pWin -> bitGravity, &nx, &ny);

      #ifdef TEST
      fprintf(stderr, "nxagentMoveCorruptedRegion: Moving the corrupted region to [%d,%d] for window [%p].\n",
                  nx, ny, (void *) pWin);
      #endif

      REGION_TRANSLATE(pWin -> pScreen, nxagentCorruptedRegion((DrawablePtr) pWin),
                           nx, ny);

      /*
       * Having moved the corrupted region, we
       * need to invalidate the pending commits
       * or otherwise the image will fall in
       * the wrong area.
       */

      nxagentValidateSplit((DrawablePtr) pWin, NULL);
    }
  }
}

Bool nxagentSynchronizeDrawableData(DrawablePtr pDrawable)
{
  int width, height, depth, length;
  unsigned int leftPad, format;

  char *data;
  DrawablePtr pSrcDrawable;
  GCPtr pGC;

  int success;

  if (pDrawable -> type == DRAWABLE_PIXMAP)
  {
    leftPad = 0;

    width  = pDrawable -> width;
    height = pDrawable -> height;
    depth  = pDrawable -> depth;

    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeDrawableData: Synchronizing drawable (%s) with geometry [%d][%d][%d].\n",
                pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window",
                    width, height, depth);
    #endif

    format = (depth == 1) ? XYPixmap : ZPixmap;

    length = nxagentImageLength(width, height, format, leftPad, depth);

    if ((data = xalloc(length)) == NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSynchronizeDrawableData: WARNING! Failed to allocate memory for the operation.\n");
      #endif

      return 0;
    }

    pSrcDrawable = (pDrawable -> type == DRAWABLE_PIXMAP ?
                        ((DrawablePtr) nxagentVirtualPixmap((PixmapPtr) pDrawable)) :
                            pDrawable);

    /*
     * Synchronize the whole pixmap if we need
     * to download a fresh copy with lossless
     * compression turned off.
     */

    if (nxagentLosslessTrap == 1)
    {
      pGC = GetScratchGC(pDrawable -> depth, pDrawable -> pScreen);

      if (pGC == NULL)
      {
        #ifdef WARNING
        fprintf(stderr, "nxagentSynchronizeDrawableData: WARNING! Failed to create the temporary GC.\n");
        #endif

        return 0;
      }

      ValidateGC(pDrawable, pGC);

      fbGetImage(pSrcDrawable, 0, 0,
                     width, height, format, AllPlanes, data);

      #ifdef TEST
      fprintf(stderr, "nxagentSynchronizeDrawableData: Forcing synchronization of "
                  "pixmap at [%p] with lossless compression.\n", (void *) pDrawable);
      #endif

      nxagentPutImage(pDrawable, pGC, depth, 0, 0,
                          width, height, leftPad, format, data);

      nxagentWaitDrawable(pDrawable);

      xfree(data);

      FreeScratchGC(pGC);

      return 1;
    }
    else if (nxagentReconnectTrap == 1)
    {
      /*
       * The pixmap data is not synchronized unless
       * we need it. We noticed we have to reconnect
       * the pixmaps referenced by the pictures used
       * to render the glyphs, and those used by the
       * GC's clip mask. The other data will be syn-
       * chronized on demand.
       */

      if (nxagentPixmapPriv((PixmapPtr) pDrawable) -> pPicture != NULL ||
              pDrawable -> depth == 1)
      {
        #ifdef TEST

        if (nxagentReconnectTrap == 1)
        {
          static int totalLength;

          totalLength += length;

          fprintf(stderr, "nxagentSynchronizeDrawableData: Reconnecting pixmap at [%p] [%dx%d] "
                      "Depth %d Size [%d]. Total is [%d].\n", (void *) pDrawable, width,
                          height, depth, length, totalLength);
        }

        #endif

        pGC = nxagentGetTempGC((PixmapPtr) pDrawable);

        if (pGC == NULL)
        {
          #ifdef WARNING
          fprintf(stderr, "nxagentSynchronizeDrawableData: WARNING! Failed to create the temporary GC.\n");
          #endif

          return 0;
        }

        fbGetImage(pSrcDrawable, 0, 0,
                       width, height, format, AllPlanes, data);

        nxagentPutImage(pDrawable, pGC, depth, 0, 0,
                            width, height, leftPad, format, data);

        nxagentWaitDrawable(pDrawable);

        xfree(data);

        return 1;
      }
      else
      {
        #ifdef TEST
        fprintf(stderr, "nxagentSynchronizeDrawableData: Skipping synchronization of "
                    "pixmap at [%p] during reconnection.\n", (void *) pDrawable);
        #endif

        nxagentMarkCorruptedRegion(pDrawable, NullRegion);

        xfree(data);

        return 1;
      }
    }

    xfree(data);
  }

  /*
   * Calling this function with the NullRegion
   * as parameter we are requesting to synchro-
   * nize the full visible corrupted region of
   * the drawable.
   */

  success = nxagentSynchronizeRegion(pDrawable, NullRegion);

  return success;
}

void nxagentSynchronizeDrawable(DrawablePtr pDrawable, int wait)
{
  pDrawable = nxagentSplitDrawable(pDrawable);

  if (nxagentLosslessTrap == 0)
  {
    if (nxagentDrawableStatus(pDrawable) == Synchronized)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSynchronizeDrawable: Not synchronizing %s [%p] with id [%ld] and "
                  "status SYNCHRONIZED.\n", (pDrawable -> type == DRAWABLE_PIXMAP ?
                      "pixmap" : "window"), (void *) pDrawable, pDrawable -> id);
      #endif

      return;
    }
  }

  /*
   * What we want here is to avoid drawing on the
   * framebuffer and just perform the operation
   * on the real X server. This is the purpose of
   * the FB trap. At the same time we also want
   * to avoid a split, so that the image will be
   * transferred in a single operation.
   */

  nxagentFBTrap = 1;

  nxagentSplitTrap = 1;

  nxagentSynchronizeDrawableData(pDrawable);

  nxagentSplitTrap = 0;

  nxagentFBTrap = 0;

  if (wait == 1 && nxagentSplitResource(pDrawable) != NULL)
  {
    nxagentWaitDrawable(pDrawable);
  }

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizeDrawable: Drawable %s [%p] with id [%ld] now synchronized.\n",
              (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"), (void *) pDrawable,
                  pDrawable -> id);
  #endif
}

int nxagentSynchronizeRegion(DrawablePtr pDrawable, RegionPtr pRegion)
{
  GCPtr pGC;
  DrawablePtr pSrcDrawable;
  BoxPtr pBox;
  RegionPtr clipRegion;

  char *data;

  int x, y;
  int width, height;
  int depth, length, format, leftPad = 0;
  int i;
  int saveTrap;
  int useExtents;
  int imageLength, corruptedLength;
  int corruptedWidth, corruptedHeight;

  #ifdef TEST
  static int totalLength;
  #endif

  if (nxagentDrawableStatus(pDrawable) == Synchronized)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeRegion: The [%s] at [%p] is already synchronized.\n",
                (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                    (void *) pDrawable);
    #endif

    return 0;
  }

  /*
   * Creating a region containing the viewable
   * area of drawable.
   */

  clipRegion = nxagentCreateRegion(pDrawable, NULL, 0, 0, pDrawable -> width, pDrawable -> height);

  /*
   * If the corrupted region is not viewable, we
   * can skip the synchronization.
   */

  REGION_INTERSECT(pDrawable -> pScreen, clipRegion, clipRegion, nxagentCorruptedRegion(pDrawable));

  if (REGION_NIL(clipRegion) == 1)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeRegion: The corrupted region [%d,%d,%d,%d] is not viewable "
                "on [%s] at [%p]. Skipping the synchronization.\n", clipRegion -> extents.x1,
                    clipRegion -> extents.y1, clipRegion -> extents.x2, clipRegion -> extents.y2,
                        (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                            (void *) pDrawable);
    #endif

    return 0;
  }

  /*
   * We can skip the synchronization if the re-
   * quested region is not corrupted. Specifying
   * a NullRegion as parameter, all the viewable
   * corrupted region will be synchronized.
   */

  if (pRegion != NullRegion)
  {
    REGION_INTERSECT(pDrawable -> pScreen, clipRegion, clipRegion, pRegion);

    if (REGION_NIL(clipRegion) == 1)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSynchronizeRegion: Region requested [%d,%d,%d,%d] already "
                  "synchronized on [%s] at [%p].\n", pRegion -> extents.x1,
                      pRegion -> extents.y1, pRegion -> extents.x2, pRegion -> extents.y2,
                          (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                              (void *) pDrawable);
      #endif

      return 0;
    }
  }

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizeRegion: Synchronizing region with coordinates [%d,%d,%d,%d] "
              "on [%s] at [%p].\n", clipRegion -> extents.x1, clipRegion -> extents.y1, 
                  clipRegion -> extents.x2, clipRegion -> extents.y2,
                      (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                          (void *) pDrawable);
  #endif

  depth  = pDrawable -> depth;

  format = (depth == 1) ? XYPixmap : ZPixmap;

  length = nxagentImageLength(pDrawable -> width, pDrawable -> height,
                                  format, leftPad, depth);

  if ((data = xalloc(length)) == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentSynchronizeRegion: WARNING! Failed to allocate memory for the operation.\n");
    #endif

    return 0;
  }

  pSrcDrawable = (pDrawable -> type == DRAWABLE_PIXMAP ?
                    ((DrawablePtr) nxagentVirtualPixmap((PixmapPtr) pDrawable)) :
                        pDrawable);

  pGC = GetScratchGC(pDrawable -> depth, pDrawable -> pScreen);

  if (pGC == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentSynchronizeRegion: WARNING! Failed to create the temporary GC.\n");
    #endif

    return 0;
  }

  ValidateGC(pDrawable, pGC);

  saveTrap = nxagentGCTrap;

  nxagentGCTrap = 0;

  nxagentFBTrap = 1;

  nxagentSplitTrap = 1;

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizeRegion: Going to synchronize [%ld] rects of [%s] at [%p].\n",
              REGION_NUM_RECTS(clipRegion), (pDrawable -> type == DRAWABLE_PIXMAP ? "Pixmap" : "Window"),
                  (void *) pDrawable);

  fprintf(stderr, "nxagentSynchronizeRegion: Extents geometry [%d,%d,%d,%d].\n",
          clipRegion -> extents.x1, clipRegion -> extents.y1, clipRegion -> extents.x2, clipRegion -> extents.y2);

  fprintf(stderr, "nxagentSynchronizeRegion: Drawable geometry [%d,%d,%d,%d].\n",
              pDrawable -> x, pDrawable -> y, pDrawable -> x - pDrawable -> width,
                  pDrawable -> y - pDrawable -> height);
  #endif

/*
FIXME: Searching for a solution to avoid multiple
       putimages of an highly 'fragmented' region.
*/
  useExtents = 0;

  corruptedWidth = clipRegion -> extents.x2 - clipRegion -> extents.x1;
  corruptedHeight = clipRegion -> extents.y2 - clipRegion -> extents.y1;

/*
FIXME: We are monitoring when and why an empty
       corrupted region can be created.
*/
  if (corruptedWidth == 0 || corruptedHeight == 0)
  {
    #ifdef WARNING

    fprintf(stderr, "nxagentSynchronizeRegion: WARNING! Trying to synchronize an empty corrupted region for %s at [%p].\n",
                (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                    (void *) pDrawable);

/*
FIXME: We don't need to keep an empty corrupted region,
       but we can't free it because this is only the
       _viewable_ corrupted region. The following output
       will help us to confirm this.
*/
    fprintf(stderr, "nxagentSynchronizeRegion: Full corrupted region extents geometry [%d,%d,%d,%d].\n",
                nxagentCorruptedRegion(pDrawable) -> extents.x1, nxagentCorruptedRegion(pDrawable) -> extents.y1,
                    nxagentCorruptedRegion(pDrawable) -> extents.x2, nxagentCorruptedRegion(pDrawable) -> extents.y2);

    #endif

    return 1;
  }

  /*
   * A drawable synchronization could be an expensive
   * operation if the corrupted region is quite frag-
   * mented. In order to keep down the number of Put-
   * Images the following optimization rules are used.
   */


  if (pDrawable -> width == corruptedWidth &&
          pDrawable -> height == corruptedHeight)
  {
    useExtents = 1;

    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeRegion: The corrupted region extents cover the full drawable.\n");
    #endif
  }
  else
  {
    imageLength = nxagentImageLength(pDrawable -> width, pDrawable -> height, format, leftPad, depth);

    corruptedLength = nxagentImageLength(corruptedWidth, corruptedHeight, format, leftPad, depth);

    if (imageLength == corruptedLength ||
            corruptedLength < 1024)
    {
      useExtents = 1;
    }

    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeRegion: Image length [%d] Corrupted region length [%d].\n",
                imageLength, corruptedLength);
    #endif
  }

  if (useExtents)
  {
    #ifdef TEST    
    fprintf(stderr, "nxagentSynchronizeRegion: Synchronizing the extents.\n");
    #endif

    fbGetImage(pSrcDrawable, clipRegion -> extents.x1, clipRegion -> extents.y1,
                   corruptedWidth, corruptedHeight, format, AllPlanes, data);

    nxagentPutImage(pDrawable, pGC, depth, clipRegion -> extents.x1, clipRegion -> extents.y1,
                        corruptedWidth, corruptedHeight, leftPad, format, data);
  }
  else
  {
    pBox = REGION_RECTS(clipRegion);

    for (i = 0; i < REGION_NUM_RECTS(clipRegion); i++)
    {
      x = pBox[i].x1;
      y = pBox[i].y1;
      width = pBox[i].x2 - pBox[i].x1;
      height = pBox[i].y2 - pBox[i].y1;

      /*
       * The box can be empty.
       */

      if (width == 0 || height == 0)
      {
        continue;
      }

      fbGetImage(pSrcDrawable, x, y, width, height, format, AllPlanes, data);

      #ifdef TEST

      length = nxagentImageLength(width, height, format, leftPad, depth);

      totalLength += length;

      fprintf(stderr, "nxagentSynchronizeRegion: Synchronizing data size [%d]. Total is [%d].\n",
                  length, totalLength);

      #endif

      nxagentPutImage(pDrawable, pGC, depth, x, y,
                          width, height, leftPad, format, data);
    }
  }

  nxagentSplitTrap = 0;

  nxagentFBTrap = 0;

  nxagentGCTrap = saveTrap;

  nxagentFreeRegion(pDrawable, clipRegion);

  xfree(data);

  FreeScratchGC(pGC);

  return 1;
}

void nxagentSynchronizeBox(DrawablePtr pDrawable, BoxPtr pBox)
{
  RegionPtr pRegion;

  if (nxagentDrawableStatus(pDrawable) == Synchronized)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeBox: The [%s] at [%p] is already synchronized.\n",
                (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                    (void *) pDrawable);
    #endif

    return;
  }

  if (pBox == NullBox)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeBox: Going to synchronize the whole [%s] at [%p].\n",
                (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                    (void *) pDrawable);
    #endif

    nxagentSynchronizeRegion(pDrawable, NullRegion);
  }
  else
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeBox: Going to create a region from box [%d,%d,%d,%d].\n",
                pBox -> x1, pBox -> y1, pBox -> x2, pBox -> y2);
    #endif

    pRegion = nxagentCreateRegion(pDrawable, NULL, pBox -> x1, pBox -> y1,
                                      pBox -> x2 - pBox -> x1, pBox -> y2 - pBox -> y1);

    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeBox: Going to synchronize the region [%d,%d,%d,%d] of "
                "[%s] at [%p].\n", pRegion -> extents.x1, pRegion -> extents.y1, pRegion -> extents.x2,
                    pRegion -> extents.y2, (pDrawable -> type == DRAWABLE_PIXMAP ? "pixmap" : "window"),
                        (void *) pDrawable);
    #endif

    nxagentSynchronizeRegion(pDrawable, pRegion);

    nxagentFreeRegion(pDrawable, pRegion);
  }
}

void nxagentSynchronizeWindowBackgroundPredicate(void *p0, XID x1, void *p2)
{
  WindowPtr pWin = (WindowPtr) p0;
  PixmapPtr pPixmap = (PixmapPtr) p2;

  XSetWindowAttributes attributes;

  if (pWin -> background.pixmap == pPixmap)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeWindowBackground: Synchronizing window [%p] with background [%p].\n",
                (void *) pWin, (void *) pWin -> background.pixmap);
    #endif

    attributes.background_pixmap = nxagentPixmap(pPixmap);

    XChangeWindowAttributes(nxagentDisplay, nxagentWindow(pWin),
                                CWBackPixmap, &attributes);

/*
FIXME: Do we really need to send a refresh
       here?
*/
    nxagentRefreshWindows(pWin);
  }
}

void nxagentSynchronizePixmapBackgroundPredicate(void *p0, XID x1, void *p2)
{
  PixmapPtr pPixmap = (PixmapPtr) p0;

  if (nxagentPixmapPriv(pPixmap) -> corruptedBackground == 1)
  {
    int i;

    /*
     * Postpone the synchronization
     * if we went out of bandwidth.
     */

    if (nxagentSynchronization == 0)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSynchronizePixmapBackgroundPredicate: Not synchronizing with bandwidth limits exceeded.\n");
      #endif

      return;
    }

    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizePixmapBackgroundPredicate: Synchronizing pixmap [%p] geometry (%dx%d).\n",
                (void *) pPixmap, pPixmap -> drawable.width, pPixmap -> drawable.height);
    #endif

    nxagentSynchronizeDrawable((DrawablePtr) pPixmap, 0);

    for (i = 0; i < currentMaxClients; i++)
    {
      if (clients[i] != NULL)
      {
        FindClientResourcesByType(clients[i], RT_WINDOW,
                                      nxagentSynchronizeWindowBackgroundPredicate, pPixmap);
      }
    }

    nxagentPixmapPriv(pPixmap) -> corruptedBackground = 0;

    nxagentCorruptedBackgrounds--;
  }
}

void nxagentSynchronizePixmapPredicate(void *p0, XID x1, void *p2)
{
  PixmapPtr pPixmap = (PixmapPtr) p0;

  if (nxagentDrawableStatus((DrawablePtr) pPixmap) == Synchronized)
  {
    return;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizePixmapPredicate: Synchronizing pixmap [%p] geometry (%dx%d).\n",
              (void *) pPixmap, pPixmap -> drawable.width, pPixmap -> drawable.height);
  #endif

  /*
   * Postpone the synchronization
   * if we went out of bandwidth.
   */

  if (nxagentSynchronization == 0)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizePixmapPredicate: Not synchronizing with bandwidth limits exceeded.\n");
    #endif

    return;
  }

  if (GetTimeInMillis() - nxagentDrawableTimestamp((DrawablePtr) pPixmap) < nxagentSynchronizationInterval)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizePixmapPredicate: Skipping the recently synchronized pixmap at [%p].\n",
                (void *) pPixmap);
    #endif

    return;
  }

  nxagentSynchronizeDrawable((DrawablePtr) pPixmap, 0);

  #ifdef TEST

  if (nxagentCheckPixmapIntegrity(pPixmap) == False ||
          nxagentDrawableStatus((DrawablePtr) pPixmap) == NotSynchronized)
  {
    fprintf(stderr, "nxagentSynchronizePixmapPredicate: WARNING! Pixmap [%p] not fully synchronized.\n",
                (void *) pPixmap);
  }

  #endif
}

void nxagentSynchronizeWindowPredicate(void *p0, XID x1, void *p2)
{
  WindowPtr pWin = (WindowPtr) p0;

  if (nxagentDrawableStatus((DrawablePtr) pWin) == Synchronized)
  {
    return;
  }

  if (nxagentWindowIsVisible(pWin) == 0)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeWindowPredicate: Skipping not visible window at [%p].\n",
                (void *) pWin);
    #endif

    return;
  }

  /*
   * Postpone the synchronization
   * if we went out of bandwidth.
   */

  if (nxagentSynchronization == 0)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeWindowPredicate: Not synchronizing with bandwidth limits exceeded.\n");
    #endif

    return;
  }

  if (GetTimeInMillis() - nxagentDrawableTimestamp((DrawablePtr) pWin) < nxagentSynchronizationInterval)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeWindowPredicate: Skipping the recently synchronized window at [%p].\n",
                (void *) pWin);
    #endif

    return;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizeWindowPredicate: Synchronizing window [%p] with geometry (%dx%d).\n",
              (void *) pWin, pWin -> drawable.width, pWin -> drawable.height);
  #endif

  nxagentSynchronizeDrawable((DrawablePtr) pWin, 0);

  #ifdef TEST

  if (nxagentDrawableStatus((DrawablePtr) pWin) == NotSynchronized)
  {
    fprintf(stderr, "nxagentSynchronizeWindowPredicate: WARNING! Window [%p] not fully synchronized.\n",
                (void *) pWin);
  }

  #endif
}

void nxagentSynchronizeAllDrawables()
{
  int i;

  if (NXQueryDisplayError(nxagentDisplay) == 1)
  {
/*
FIXME: All drawables should be set as synchronized and
       never marked as corrupted while the display is
       down.
*/
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeAllDrawables: WARNING! Not synchronizing the drawables "
                "with the display down.\n");
    #endif

    return;
  }

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizeAllDrawables: Synchronizing [%d] windows [%d] pixmaps "
              "and [%d] backgrounds.\n", nxagentCorruptedWindows, nxagentCorruptedPixmaps,
                  nxagentCorruptedBackgrounds);
  #endif

  for (i = 0; i < currentMaxClients &&
           nxagentCorruptedWindows > 0 &&
               nxagentSynchronization == 1; i++)
  {
    if (clients[i] != NULL)
    {
      FindClientResourcesByType(clients[i], RT_WINDOW,
                                    nxagentSynchronizeWindowPredicate, NULL);
    }
  }

  /*
   * Synchronize the pixmaps used as window
   * backgrounds.
   */

  for (i = 0; i < currentMaxClients &&
           nxagentCorruptedBackgrounds > 0 &&
               nxagentSynchronization == 1; i++)
  {
    /*
     * The RT_NX_PIXMAP resource type is only
     * assigned to the server client, so we
     * don't need to iterate through the other
     * clients.
     */

    if (i == serverClient -> index)
    {
      FindClientResourcesByType(clients[i], RT_NX_PIXMAP,
                                    nxagentSynchronizePixmapBackgroundPredicate, NULL);
    }

    if (clients[i] != NULL)
    {
      FindClientResourcesByType(clients[i], RT_PIXMAP,
                                    nxagentSynchronizePixmapBackgroundPredicate, NULL);
    }
  }

/*
FIXME: We don't really need to synchronize
       the pixmaps.
*/
#if 0
  /*
   * If there is bandwidth remaining, synchronize
   * the pixmaps. Synchronizing a pixmap doesn't
   * produce any visible results. Better is to
   * synchronize them on demand, before using the
   * pixmap in a copy or in a composite operation.
   */

  for (i = 0; i < currentMaxClients &&
           nxagentCorruptedPixmaps > 0 &&
               nxagentSynchronization == 1; i++)
  {
    if (i == serverClient -> index)
    {
      FindClientResourcesByType(clients[i], RT_NX_PIXMAP,
                                    nxagentSynchronizePixmapPredicate, NULL);
    }

    if (clients[i] != NULL)
    {
      FindClientResourcesByType(clients[i], RT_PIXMAP,
                                    nxagentSynchronizePixmapPredicate, NULL);
    }
  }
#endif

  #ifdef TEST
  fprintf(stderr, "nxagentSynchronizeAllDrawables: There are now [%d] windows [%d] pixmaps "
              "and [%d] backgrounds to synchronize.\n", nxagentCorruptedWindows,
                  nxagentCorruptedPixmaps, nxagentCorruptedBackgrounds);
  #endif
}

void nxagentSynchronizeShmPixmap(DrawablePtr pDrawable, int xPict, int yPict,
                                     int wPict, int hPict)
{
  GCPtr pGC;
  char *data;
  int width, height;
  int depth, length, format;
  CARD32 attributes[3];

  int saveTrap;

  if (pDrawable -> type == DRAWABLE_PIXMAP &&
         nxagentIsShmPixmap((PixmapPtr) pDrawable) == 1)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSynchronizeShmPixmap: WARNING! Synchronizing shared pixmap at [%p].\n",
                (void *) pDrawable);
    #endif

    pGC = GetScratchGC(pDrawable -> depth, pDrawable -> pScreen);

    attributes[0] = 0x228b22;
    attributes[1] = 0xffffff;
    attributes[2] = FillSolid;

    ChangeGC(pGC, GCForeground | GCBackground | GCFillStyle, attributes);

    ValidateGC(pDrawable, pGC);

    width  = (wPict != 0 && wPict <= pDrawable -> width) ? wPict : pDrawable -> width;
    height = (hPict != 0 && hPict <= pDrawable -> height) ? hPict : pDrawable -> height;

    depth  = pDrawable -> depth;

    format = (depth == 1) ? XYPixmap : ZPixmap;

    length = nxagentImageLength(width, height, format, 0, depth);

    saveTrap = nxagentGCTrap;

    nxagentGCTrap = 0;

    nxagentSplitTrap = 1;

    nxagentFBTrap = 1;

    if ((data = xalloc(length)) != NULL)
    {
      fbGetImage(nxagentVirtualDrawable(pDrawable), xPict, yPict,
                     width, height, format, 0xffffffff, data);

      nxagentPutImage(pDrawable, pGC, depth, xPict, yPict,
                          width, height, 0, format, data);

      xfree(data);
    }
    #ifdef WARNING
    else
    {
      fprintf(stderr, "nxagentSynchronizeShmPixmap: WARNING! Failed to allocate memory for the operation.\n");
    }
    #endif

    nxagentGCTrap = saveTrap;

    nxagentSplitTrap = 0;

    nxagentFBTrap = 0;

    FreeScratchGC(pGC);
  }
}

#ifdef DUMP

/*
 * This function is useful to visualize a pixmap and check
 * its data consistency. To avoid the creation of many
 * windows, one pixmap only can be monitored at a time.
 */

Bool nxagentPixmapOnShadowDisplay(PixmapPtr pMap)
{
  static Display *shadow;
  static Window win;
  static int init = True;
  static int showTime;
  static PixmapPtr pPixmap;
  static int depth;
  static int width;
  static int height;
  static int length;
  static unsigned int format;

  XlibGC gc;
  XGCValues value;
  XImage *image;
  Visual *pVisual;
  char *data = NULL;


  if (init)
  {
    if (pMap == NULL)
    {
      return False;
    }
    else
    {
      pPixmap = pMap;
    }

    depth = pPixmap -> drawable.depth;
    width = pPixmap -> drawable.width;
    height = pPixmap -> drawable.height;
    format = (depth == 1) ? XYPixmap : ZPixmap;

    shadow = XOpenDisplay("localhost:0");

    if (shadow == NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentPixmapOnShadowDisplay: WARNING! Shadow display not opened.\n");
      #endif

      return False;
    }

    init = False;

    win = XCreateSimpleWindow(shadow, DefaultRootWindow(shadow), 0, 0,
                                  width, height, 0, 0xFFCC33, 0xFF);

    XMapWindow(shadow, win);
    XClearWindow(shadow, win);
  }

/*
FIXME: If the pixmap has a different depth from the window, the
       XPutImage returns a BadMatch. For example this may happens if
       the Render extension is enabled.
       Can we fix this creating a new pixmap?
*/

  if (DisplayPlanes(shadow, DefaultScreen(shadow)) != depth)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentPixmapOnShadowDisplay: Pixmap and Window depths [%d - %d] are not equals!\n",
                depth, DisplayPlanes(shadow, DefaultScreen(shadow)));
    #endif

    return False;
  }

  /*
   * If the framebuffer is updated continuously, the nxagent
   * visualization become too much slow.
   */

  if ((GetTimeInMillis() - showTime) < 500)
  {
    return False;
  }

  showTime = GetTimeInMillis();

  length = nxagentImageLength(width, height, format, 0, depth);

  if ((data = xalloc(length)) == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentPixmapOnShadowDisplay: WARNING! Failed to allocate memory for the operation.\n");
    #endif

    return False;
  }

  fbGetImage((DrawablePtr) nxagentVirtualPixmap(pPixmap), 0, 0,
                 width, height, format, AllPlanes, data);

  pVisual = nxagentImageVisual((DrawablePtr) pPixmap, depth);

  image = XCreateImage(nxagentDisplay, pVisual,
                           depth, format, 0, (char *) data,
                               width, height, BitmapPad(nxagentDisplay),
                                   nxagentImagePad(width, format, 0, depth));

  if (image == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentPixmapOnShadowDisplay: XCreateImage failed.\n");
    #endif

    if (data != NULL)
    {
      xfree(data);
    }

    return False;
  }

  value.foreground = 0xff0000;
  value.background = 0x000000;
  value.plane_mask = 0xffffff;
  value.fill_style = FillSolid;

  gc = XCreateGC(shadow, win, GCBackground |
                     GCForeground | GCFillStyle | GCPlaneMask, &value);

  NXCleanImage(image);

  XPutImage(shadow, win, gc, image, 0, 0, 0, 0, width, height);

  XFreeGC(shadow, gc);

  if (image != NULL)
  {
    XDestroyImage(image);
  }

  return True;
}

Bool nxagentFbOnShadowDisplay()
{
  static Display *shadow;
  static Window win;
  static int init = True;
  static int showTime;
  static int prevWidth, prevHeight;

  XlibGC gc;
  XGCValues value;
  XImage *image;
  Visual *pVisual;
  WindowPtr pWin = WindowTable[0];
  unsigned int format;
  int depth, width, height, length;
  char *data = NULL;


  if (pWin == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentFbOnShadowDisplay: The parent window is NULL.\n");
    #endif

    return False;
  }

  depth = pWin -> drawable.depth;
  width = pWin -> drawable.width;
  height = pWin -> drawable.height;
  format = (depth == 1) ? XYPixmap : ZPixmap;

  if (init)
  {
    shadow = XOpenDisplay("localhost:0");

    if (shadow == NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentFbOnShadowDisplay: WARNING! Shadow display not opened.\n");
      #endif

      return False;
    }

    init = False;

    prevWidth = width;
    prevHeight = height;

    win = XCreateSimpleWindow(shadow, DefaultRootWindow(shadow), 0, 0,
                                  width, height, 0, 0xFFCC33, 0xFF);

    XMapWindow(shadow, win);
    XClearWindow(shadow, win);
  }

  if (DisplayPlanes(shadow, DefaultScreen(shadow)) != depth)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentFbOnShadowDisplay: Depths [%d - %d] are not equals!\n",
                depth, DisplayPlanes(shadow, DefaultScreen(shadow)));
    #endif

    return False;
  }

  /*
   * If the framebuffer is updated continuously, the nxagent
   * visualization becomes too much slow.
   */

  if ((GetTimeInMillis() - showTime) < 500)
  {
    return False;
  }

  showTime = GetTimeInMillis();

  /*
   * If the root window is resized, also the window on shadow
   * display must be resized.
   */

  if (prevWidth != width || prevHeight != height)
  {
    XWindowChanges values;

    prevWidth = width;
    prevHeight = height;

    values.width = width;
    values.height = height;
    XConfigureWindow(shadow, win, CWWidth | CWHeight, &values);
  }

  length = nxagentImageLength(width, height, format, 0, depth);

  if ((data = xalloc(length)) == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentFbOnShadowDisplay: WARNING! Failed to allocate memory for the operation.\n");
    #endif

    return False;
  }

  fbGetImage((DrawablePtr)pWin, 0, 0,
                 width, height, format, AllPlanes, data);

  pVisual = nxagentImageVisual((DrawablePtr) pWin, depth);

  image = XCreateImage(nxagentDisplay, pVisual,
                           depth, format, 0, (char *) data,
                               width, height, BitmapPad(nxagentDisplay),
                                   nxagentImagePad(width, format, 0, depth));

  if (image == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentFbOnShadowDisplay: XCreateImage failed.\n");
    #endif

    if (data)
    {
      xfree(data);
    }

    return False;
  }

  value.foreground = 0xff0000;
  value.background = 0x000000;
  value.plane_mask = 0xffffff;
  value.fill_style = FillSolid;

  gc = XCreateGC(shadow, win, GCBackground |
                     GCForeground | GCFillStyle | GCPlaneMask, &value);

  NXCleanImage(image);

  XPutImage(shadow, win, gc, image, 0, 0, 0, 0, width, height);

  XFreeGC(shadow, gc);

  if (image != NULL)
  {
    XDestroyImage(image);
  }

  return True;
}

#endif

#ifdef TEST

void nxagentPrintResourceTypes()
{
  fprintf(stderr, "nxagentPrintResourceTypes: RT_PIXMAP [%lu].\n", (unsigned long) RT_PIXMAP);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_NX_PIXMAP [%lu].\n", (unsigned long) RT_NX_PIXMAP);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_GC [%lu].\n", (unsigned long) RT_GC);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_NX_GC [%lu].\n", (unsigned long) RT_NX_GC);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_FONT [%lu].\n", (unsigned long) RT_FONT);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_NX_FONT [%lu].\n", (unsigned long) RT_NX_FONT);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_CURSOR [%lu].\n", (unsigned long) RT_CURSOR);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_WINDOW [%lu].\n", (unsigned long) RT_WINDOW);
  fprintf(stderr, "nxagentPrintResourceTypes: RT_COLORMAP [%lu].\n", (unsigned long) RT_COLORMAP);
}

void nxagentPrintResourcePredicate(void *value, XID id, XID type, void *cdata)
{
  fprintf(stderr, "nxagentPrintResourcePredicate: Resource [%p] id [%lu] type [%lu].\n",
              (void *) value, (unsigned long) id, (unsigned long) type);
}

void nxagentPrintResources()
{
  Bool result;
  int i;

  nxagentPrintResourceTypes();

  for (i = 0; i < MAXCLIENTS; i++)
  {
    if (clients[i])
    {
      fprintf(stderr, "nxagentPrintResources: Printing resources for client [%d]:\n",
                  i);

      FindAllClientResources(clients[i], nxagentPrintResourcePredicate, &result);
    }
  }
}

#endif


